///////////////////////////////////////////////////////////////////////////////////////////////
// drvapitest.cpp : Test module for the DRvApi
// API for Datatec Rentas Vitalicias
///////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include "../drvapi/drvapi.h"

/////////////////////////////////////////////////////////////////////////////////////////////////

CDrvApi D_RV_Api_Obj;

#define BUYER_SIDE         1
#define SELLER_SIDE        2
#define BUYER_SELLER_SIDE  3
#define STATUS_CHANGE      4

char *L_status[] = {"N/A",
                    "Rejected",
                    "Accepted",
                    "Sent To Customer",
                    "Viewed",
                    "Time Out Accept",
                    "Time Out Reject",
                    "Overridden",
                    };
static int L_dapi_connected;
static int L_counter;

/////////////////////////////////////////////////////////////////////////////////////////////////

static size_t   L_get_line                    (char *line, int max_len);
static void     Test_SendBackGuibosMsg        (void);
static void     Test_SendFITradeMsg           (void);
static int      Test_Get_MsgType_MsgNumber    (char *mess_type, unsigned long *mess_number);
static int      Test_Get_Username_Password    (char *username_value, char *password_value);
static void     DoChangeStatusProcess         (void);

/////////////////////////////////////////////////////////////////////////////////////////////////

#ifndef WIN32
#include <stdlib.h>
#include <termios.h>
#include <unistd.h>

static struct termios initial_settings, new_settings;
static int peek_character = -1;

static void init_keyboard();
static void close_keyboard();
static int  _kbhit();
static int  _getch();
#else
#define stricmp _stricmp
#endif


/////////////////////////////////////////////////////////////////////////////////////////////////

int main(int argc, char * argv[])
{
  char           line[300];
  char           username[256];
  char           password[256];
  char           vr_dapi  [16],
                 comp_type[16];
  unsigned long  result;
  char           messtype[256];
  unsigned long  messnum;
  
#ifndef WIN32
  init_keyboard();
#endif

  // D_RV_Api_Obj.ProcessMyEvents ();
  for (;;)
  {
    D_RV_Api_Obj.GetVersion(vr_dapi);
    D_RV_Api_Obj.GetCompilationType(comp_type);
    sprintf(messtype, "\nSimple D_RV_API test - Version %s %s\n", vr_dapi, comp_type);

    // Ask for something to do
    printf(messtype);
    printf("-------------------------------------------\n\n"
              "Options are:\n\n"
              "L- Make login.\n"
              "D- set(1)-unset(0) mode to Display/process msgs as they are received (%d).\n"
              "V- View fields depth.\n"
              "P- set msg as Processed.\n"
              "E- send FI TradE msg.\n"
              "G- Send back Guibos message.\n"
              "O- make logOut.\n"
              "Q- Quit.\n\n"
              "Option ? \n",
              D_RV_Api_Obj.Verbouse_Mode());

    if (!L_get_line (line, sizeof (line)))
      continue;
        
    switch (toupper(line[0]))
    {
    case 'D':
        D_RV_Api_Obj.Set_Verbose_Mode();
        break;
    case 'E':
      // make a message and send
      Test_SendFITradeMsg();
      break;
    case 'G':
        Test_SendBackGuibosMsg();
        break;
    case 'L':
      if (Test_Get_Username_Password (username, password))
        break;
      // make a login 
      result = D_RV_Api_Obj.Login (username, password);
      if (result != 0)
      {
        printf ("Error code %ld\n", result);
        break;
      }

      // flag que indica que el dapi ya fue conectado.
      L_dapi_connected = 1;
      break;
    case 'V':
      // View the fields values of a message.
      if (Test_Get_MsgType_MsgNumber (messtype, &messnum))
        break;
      D_RV_Api_Obj.DisplayMessageContent (messtype, messnum);
      break;
    case 'P':
      // flag a message as processed
      // the message will not be received any more in a case of reconnection
      if (Test_Get_MsgType_MsgNumber (messtype, &messnum))
        break;

      // the message will not send any more in a case of reconnection
      D_RV_Api_Obj.FlagMessage (messtype, messnum);
      break;
    case 'O':
      // disconnect and do all again
      D_RV_Api_Obj.Logout ();
      L_dapi_connected = 0;
      break;
    case 'Q':
      // leave it
      goto FINISH;
    }
  }

FINISH:
  // Disconnect and finnish
  D_RV_Api_Obj.Logout ();

  printf ("Interactive testing finalized\n");

#ifndef WIN32
  close_keyboard();
#endif

  return 0;
}

void Test_SendBackGuibosMsg (void)
{
  char           messtype[256];
  unsigned long  messnum;
  char           ch_side[2];
  int            side = BUYER_SELLER_SIDE;

  // Get the message type and number.
  if (Test_Get_MsgType_MsgNumber (messtype, &messnum))
    return;

  if (stricmp(messtype, L_GuibosType))
  {
    printf ("The message type must be Guibos one\n");
    return;
  }

  printf ("Select buyer and/or seller data filled option:\n"
          "1- Buyer Side.\n"
          "2- Seller Side.\n"
          "3- Buyer and Seller Side.\n"
          "4- Status.\n"
          "5- Login Conditions.\n\n"
          "Choose one Option ? \n");

  if (!L_get_line (ch_side, sizeof (ch_side)))
    return;

  // Se convierte a numero el caracter.
  side = atoi(ch_side);

  if (!strcmp (messtype, L_GuibosType))
  {
    
    if (side == 5)
    {
      if (L_dapi_connected)
      {
        printf ("Login already was setting. Please press logout to set the login conditions.\n");
        return;
      }

      D_RV_Api_Obj.AddLoginCondition (L_GuibosType, "instrument",    "==",   "SWAP");
      D_RV_Api_Obj.AddLoginCondition (L_GuibosType, "instrument",    "==",   "WDF");
      D_RV_Api_Obj.AddLoginCondition (L_GuibosType, "|",             "", "");
      D_RV_Api_Obj.AddLoginCondition (L_GuibosType, "buyer_dealer",  "LIKE", "AAAL%");
      D_RV_Api_Obj.AddLoginCondition (L_GuibosType, "seller_dealer", "LIKE", "BBBL%");
      D_RV_Api_Obj.AddLoginCondition (L_GuibosType, "|",             "", "");
      D_RV_Api_Obj.AddLoginCondition (L_GuibosType, "&",             "", "");
      printf ("AddLoginCondition test processed.\n");
      return;
    }

    if (D_RV_Api_Obj.TakeMessage (messtype, messnum))
      return;

    D_RV_Api_Obj.InitMessage (""); // "" => initiates the same taken messages with it's content
                             // this avoids to construct all the message fields
    // Send the same message received with buyer and seller data filled.
    if (side & BUYER_SIDE)
    {
      D_RV_Api_Obj.FillBuyerFields();
      printf ("Buyer fields was filled.\n");
    }

    if (side & SELLER_SIDE)
    {
      D_RV_Api_Obj.FillSellerFields();
      printf ("Seller fields was filled.\n");
    }
    
    if (side == BUYER_SELLER_SIDE)
    {
      D_RV_Api_Obj.SetFieldBSTR(L_GuibosType, "flag_side_changed", "A");//"B" - Buyer,  "S" - Seller, "A" - both sides
      printf ("Buyer and Seller fields were filled.\n");
    }

    if (side == STATUS_CHANGE)
    {
      DoChangeStatusProcess ();
      printf ("Status fields were filled.\n");
    }

    // send the message
    printf ("Message Guibos DRN test processed.\n");
    D_RV_Api_Obj.SendMessage(L_GuibosType);

  }
  else if (!strcmp (messtype, L_GuibosWhType))
  {
    D_RV_Api_Obj.TakeMessage (messtype, messnum);

    D_RV_Api_Obj.InitMessage (""); // "" => initiates the same taken messages with it's content
                              // this avoids to construct all the message fields
    D_RV_Api_Obj.SetFieldBSTR(L_GuibosWhType, "published_field",              "");//<=
    D_RV_Api_Obj.SetFieldBSTR(L_GuibosWhType, "guibos_drn",                   "KKPLZNH1");//<=
    D_RV_Api_Obj.SetFieldBSTR(L_GuibosWhType, "guibos_version",               "1.234");//<=
    D_RV_Api_Obj.SetFieldBSTR(L_GuibosWhType, "guibos_processed_hour",        "150430");//<=
    D_RV_Api_Obj.SetFieldBSTR(L_GuibosWhType, "guibos_processed_date",        "20031009");//<=
    D_RV_Api_Obj.SetFieldBSTR(L_GuibosWhType, "guibos_status_error",          "");//<=

    printf ("Message GuibosWh DGapi DRN test processed.\n");
    D_RV_Api_Obj.SendMessage (L_GuibosWhType);
  }
}

void Test_SendFITradeMsg (void)
{
  char            msg_type[256];
  unsigned long   msg_id;
  int             int_order_mercado;
  int             field_type;
  char            field_val[64];
  char            str_aux[32];
  char            order_mercado[64];
  char            order_nombre_titulo[64];
  unsigned short  order_liquidacion;
  unsigned short  order_concertacion;
  short           order_xdias;
  unsigned short  order_dias;
  char            order_fecha[64];
  char            order_hora[64];
  char            order_orden_operacion[64];
  char            order_tipo_operacion[64];
  char            order_precio[64];
  char            order_numero_transado[64];
  char            order_o_d[64];
  char            order_bandera_broker_postura[64];
  char            order_bandera_broker_postura_1[64];
  char            order_orden_compra_venta[64];
  char            order_agresor_comprador_vendedor[64];
  char            order_nombre_comprador[64];
  unsigned short  order_codigo_comprador;
  char            order_nombre_vendedor[64];
  unsigned short  order_codigo_vendedor;

  // Get the message type and number.
  if (Test_Get_MsgType_MsgNumber (msg_type, &msg_id))
    return;

  // Get the message selected.
  if (D_RV_Api_Obj.TakeMessage (msg_type, msg_id))
    return;

  // En este ejemplo no hay backoffice para tener el proceso completo de una transaccin.
  // Como ejemplo, lo que se hace es tomar una transaccin realizada
  // en el Sistema para enviar de regreso al Sistema el hecho definitivo, usando
  // los mismos datos de la transaccin Datatec.
  // Sin embargo, esto no es una restriccin. La aplicacin que usa el DAPI puede enviar
  // ms mensajes, si es que el mdulo de asignacin del Backoffice gener ms hechos.
  if (stricmp(msg_type, L_Orden_TransaccionType)) 
  {
    printf ("Selected record is NOT a trade (MsgType=Orden_Transaccion)(Field clase_mensaje=5).");
    return;
  }
  if (!D_RV_Api_Obj.GetField("clase_mensaje", field_val) ||
      memcmp(field_val, " 5", 2))
  {
    printf ("Selected record is NOT a trade (Field clase_mensaje=5).");
    return;
  }

  field_type = D_RV_Api_Obj.GetField("mercado", order_mercado);
  int_order_mercado = atoi(order_mercado);

  // The selected record must be of Fixed Income markets 69, 140, 141.
  if (int_order_mercado != 69 && int_order_mercado != 140 && int_order_mercado != 141)
  {
    printf ("Selected record is NOT a trade of Fixed Income markets 69,140,141 (Field mercado).");
    return;
  }

  ///////////////////////////////////////////////////////////////////////////
  // En las siguientes lineas se obtienen los datos del instrumento y de
  // la transaccion.
  // Esto se lo hace porque el ejemplo no tiene los procesos de asignacion
  // del backoffice, y para simular el envio se necesitan estos datos.
  ///////////////////////////////////////////////////////////////////////////
  // Datos del instrumento.
  field_type = D_RV_Api_Obj.GetField("nombre_titulo", order_nombre_titulo);
  field_type = D_RV_Api_Obj.GetField("liquidacion", field_val);
  order_liquidacion = *((unsigned short *)field_val);
  field_type = D_RV_Api_Obj.GetField("concertacion", field_val);
  order_concertacion = *((unsigned short *)field_val);
  field_type = D_RV_Api_Obj.GetField("xdias", field_val);
  order_xdias = *((short *)field_val);
  field_type = D_RV_Api_Obj.GetField("dias", field_val);
  order_dias = *((unsigned short *)field_val);
  // Datos de la transaccion.
  field_type = D_RV_Api_Obj.GetField("fecha", order_fecha);
  field_type = D_RV_Api_Obj.GetField("hora", order_hora);
  field_type = D_RV_Api_Obj.GetField("orden_operacion", order_orden_operacion);
  field_type = D_RV_Api_Obj.GetField("tipo_operacion", order_tipo_operacion);
  field_type = D_RV_Api_Obj.GetField("precio", order_precio);
  field_type = D_RV_Api_Obj.GetField("numero_transado", order_numero_transado);
  field_type = D_RV_Api_Obj.GetField("o_d", order_o_d);
  field_type = D_RV_Api_Obj.GetField("bandera_broker_postura", order_bandera_broker_postura);
  field_type = D_RV_Api_Obj.GetField("bandera_broker_postura_1", order_bandera_broker_postura_1);

  // Como el ejemplo no tiene los procesos del backoffice, se toman los siguientes
  // datos del mismo mensaje de transaccin.
  //  - cdigo y nombre comprador
  //  - cdigo y nombre vendedor
  //  - Operacin fue compra o venta
  //  - Quien fue el agresor.
  // En este ejemplo, los datos del comprador y vendedor pueden estar vacos cuando la
  // transaccion fue realizada por un broker sin especificar los participantes.
  // Sin embargo, el backoffice si debera tener estos datos luego de
  // hacer el proceso de asignacin.
  if (*order_o_d == 'D')
  {
    strcpy(order_orden_compra_venta, "C");
    // Agresor es el vendedor.
    strcpy(order_agresor_comprador_vendedor, "CE");

    field_type = D_RV_Api_Obj.GetField("numero_sucursal_postura", field_val);
    order_codigo_comprador = *((unsigned short *)field_val);
    field_type = D_RV_Api_Obj.GetField("numero_sucursal_postura_1", field_val);
    order_codigo_vendedor = *((unsigned short *)field_val);

    if (*order_bandera_broker_postura == 'S')
    {
      // Order entered by a broker
      field_type = D_RV_Api_Obj.GetField("sucursal_4c_banco_postura", order_nombre_comprador);
    }
    else
    {
      field_type = D_RV_Api_Obj.GetField("sucursal_4c_postura", order_nombre_comprador);
    }

    if (*order_bandera_broker_postura_1 == 'S')
    {
      // Order aggressed by a broker
      field_type = D_RV_Api_Obj.GetField("sucursal_4c_banco_postura_1", order_nombre_vendedor);
    }
    else
    {
      field_type = D_RV_Api_Obj.GetField("sucursal_4c_postura_1", order_nombre_vendedor);
    }
  }
  else  // order_o_d = "O"
  {
    strcpy(order_orden_compra_venta, "V");
    // Agresor es el comprador.
    strcpy(order_agresor_comprador_vendedor, "CO");
    field_type = D_RV_Api_Obj.GetField("numero_sucursal_postura", field_val);
    order_codigo_vendedor = *((unsigned short *)field_val);
    field_type = D_RV_Api_Obj.GetField("numero_sucursal_postura_1", field_val);
    order_codigo_comprador = *((unsigned short *)field_val);

    if (*order_bandera_broker_postura == 'S')
    {
      // Order entered by a broker
      field_type = D_RV_Api_Obj.GetField("sucursal_4c_banco_postura", order_nombre_vendedor);
    }
    else
    {
      field_type = D_RV_Api_Obj.GetField("sucursal_4c_postura", order_nombre_vendedor);
    }

    if (*order_bandera_broker_postura_1 == 'S')
    {
      // Order aggressed by a broker
      field_type = D_RV_Api_Obj.GetField("sucursal_4c_banco_postura_1", order_nombre_comprador);
    }
    else
    {
      field_type = D_RV_Api_Obj.GetField("sucursal_4c_postura_1", order_nombre_comprador);
    }
  }

  ///////////////////////////////////////////////////////////////////////////
  // De aqu en adelante es el proceso que se tendra que realizar usando
  // los valores que tenga el backoffice.
  // El uso de la funcin PrintText es solamente para desplegar en la ventana
  // de Windows. Las funciones que se deben usar son:
  //  InitMessage("Orden_Transaccion")
  //  SetField(...)
  //  SetField(...)
  //  ...
  //  SendMessage()
  ///////////////////////////////////////////////////////////////////////////

  // Inicializar el mensaje para ser enviado
  // Se inicializa el mensaje para tener un mensaje nuevo y sin campos llenados.
  D_RV_Api_Obj.InitMessage(L_Orden_TransaccionType);

  // 
  // Debe ser definido con 71 (alfanumerico de 2)
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "clase_mensaje", "71");

  // S - Le indica al DS que este mensaje no se debe grabar en el log de los destinatarios del mensaje
  // (alfanumerico de 1)
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "no_guardar_log_destinatarios", "S");

  // Es el tipo de operacin (alfanumerico de 1), sus posibles valores son:
  // A = Alta, 
  // B = Baja
  // M = Modificacin
  //Como ejemplo, se estn poniendo todos los mensajes como Alta.
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "bandera", "A");

  // El numero del mercado a que corresponde este mensaje (alfanumerico de 3)
  // Los posibles cdigos de mercado son: 69, 140  141
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "mercado", order_mercado);

  // La fecha de la transaccion. Formato: AAAAMMDD. (alfanumerico de 8)
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "fecha_postura", order_fecha);

  // Es la hora_hecho, es decir es la hora de la ultima operacion sobre la postura
  // (alfanumerico de 6)
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "hora_operacion", order_hora);

  // Folio.  El numero de papeles que estan asignados (entero doble)
  // Como ejemplo, se est usando un contador.
  L_counter++;
  sprintf (str_aux, "%d", L_counter);
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "numero_asignado", str_aux);

  // Cdigo alterno del grupo, usado para ingresar una nueva lnea (alfanumerico de 1)
  // Los posibles valores son: C  V
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "orden_operacion", order_orden_operacion);

  // Es el tipo de operacin (alfanumerico de 2)
  // Los posibles valores son:
  // R = Reporto
  // D = Directa
  // DR  = Directa y repoto
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "tipo_operacion", order_tipo_operacion);

  // Nombre del ttulo (alfanumerico de 10). Ejemplo: JL09, DC10, CTSI, G2I, etc.
  // El campo usado para el mensaje es texto_term, pero el dato que se enva es
  // el nombre del ttulo.
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "texto_term", order_nombre_titulo);

  // Horas hasta la liquidacin (entero sin signo), posibles valores:
  // 0 = Mismo da (MD)
  // 24, 48, 72, 96 horas
  sprintf (str_aux, "%hu", order_liquidacion);
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "liquidacion", str_aux);

  // Concertacion (entero sin signo). Los posibles valores son:
  // 0 = Orden en firme
  // 1 = Cama
  // 2 = Ronda
  // 3 = Chinga
  // 4 = Montos
  // 5 = Ponderada
  sprintf (str_aux, "%hu", order_concertacion);
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "concertacion", str_aux);

  // Plazo de la transaccin  plazo inferior cuando es un rango, ejemplo (60-360) 
  // (entero)
  sprintf (str_aux, "%hd", order_xdias);
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "xdias", str_aux);

  // Plazo superior cuando es un rango, ejemplo (60-360) 
  // (entero si signo)
  sprintf (str_aux, "%hu", order_dias);
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "dias", str_aux);

  // Cdigo de la institucin compradora (alfanumerico de 10). Ejmplo: BANAMEX
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "referencia", order_nombre_comprador);

  // Nmero asociado a la institucin compradora (entero sin signo). Ejemplo: 26
  sprintf (str_aux, "%hu", order_codigo_comprador);
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "numero_sucursal_postura", str_aux);

  // Nmero asociado a la institucin vendedora (entero sin signo). Ejemplo: 54
  sprintf (str_aux, "%hu", order_codigo_vendedor);
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "numero_sucursal_postura_1", str_aux);

  // Cdigo de la institucin vendedora (alfanumerico de 10). Ejemplo: INDELPRO
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "referencia_titulo", order_nombre_vendedor);

  // Precio de la transaccin (numerico 13 enteros, 4 decimales). Ejemplo: 7.3, 5.21, etc
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "precio", order_precio);

  // Monto de la transaccin (numerico 13 enteros y 2 decimales)
  // El monto se envia en unidades a partir de versin 6.14f Sipo Corro.
  // Antes se enviaba en millones, pero se hizo el cambio para que se usen
  // las mismas unidades que en el Sistema.
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "numero_transado", order_numero_transado);

  // Los posibles valores son: 
  // C = orden/pedido de compra
  // V = orden/pedido de venta
  // (alfanumerico de 1)
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "o_d", order_orden_compra_venta);

  // Indica el agresor (alfanumerico de 3), los posibles valores son: 
  // CO = Comprador fue el agresor,  
  // CE = Vendedor fue el agresor
  D_RV_Api_Obj.SetFieldBSTR(L_Orden_TransaccionType, "entidad_publica_3c", order_agresor_comprador_vendedor);

  // Enviar el mensaje
  // El mtodo SendMessage() enva por omisin el mensaje a los DMservers.
  // El DMserver con control de estadsticas procesa este mensaje e ingresa
  // los registros que son vistos en la ventana de hechos del Sistema.
  D_RV_Api_Obj.SendMessage(L_Orden_TransaccionType);
}

int Test_Get_MsgType_MsgNumber (char *mess_type, unsigned long *mess_number)
{
  char          value1_str[256+1];
  char          value2_str[256+1];
  unsigned long message_number_ulong;

  if (!mess_type || !mess_number)
    return 1;

  *mess_type = 0;
  *mess_number = 0;

  printf ("Type the message type (for example, Orden_Transaccion):\n");
  if (!L_get_line (value1_str, sizeof (value1_str)))
    return 2;

  printf ("Type the message number (for example, 123):\n");
  if (!L_get_line (value2_str, sizeof (value2_str)))
    return 3;

  if (!sscanf (value2_str, "%lu", &message_number_ulong))
  {
    printf ("Invalid message number, try again.\n");
    return 4;
  }

  strcpy (mess_type, value1_str);
  *mess_number = message_number_ulong;

  return 0;
}

int Test_Get_Username_Password (char *username_value, char* password_value)
{
  char          value1_str[256+1];
  char          value2_str[256+1];

  if (!username_value || !password_value)
    return 1;

  *username_value = 0;
  *password_value = 0;

  // Ask for the username and password
  printf ("Type the username (ENTER to use value from dapi.cfg):\n");
  if (L_get_line (value1_str, sizeof (value1_str)))
    strcpy (username_value, value1_str);

  printf ("Type the password (ENTER to use value from dapi.cfg):\n");
  if (L_get_line (value2_str, sizeof (value2_str)))
    strcpy (password_value, value2_str);

  return 0;
}

void DoChangeStatusProcess (void)
{
  char  text_reason[400];
  char  field_val  [256];

  D_RV_Api_Obj.GetFieldBSTR ("trade_identification", field_val);
  D_RV_Api_Obj.SetFieldBSTR(L_GuibosType, "trade_identification", field_val);

  D_RV_Api_Obj.GetFieldBSTR ("instance_trade", field_val);
  D_RV_Api_Obj.SetFieldBSTR(L_GuibosType, "instance_trade", field_val);

  D_RV_Api_Obj.SetFieldBSTR(L_GuibosType, "bandera", "S"); // status changed REQUIRED
  D_RV_Api_Obj.SetFieldBSTR(L_GuibosType, "flag_side_changed", "A");
  //; B  The BUYER STP STATUS IS specified    
  //; S  The SELLER STP STATUS IS specified    
  //; A  Both BUYER and SELLER are specified

  text_reason[0] = 0;
  strcat (text_reason, "Some reason to explain ");

  //; Text with the reason for the BUYER
  D_RV_Api_Obj.SetFieldBSTR(L_GuibosType, "buyer_side_status_error", text_reason);
  
  //; Text with the reason for the SELLER
  D_RV_Api_Obj.SetFieldBSTR(L_GuibosType, "seller_side_status_error", text_reason);
  
  //; Status of the BUYERS transaction
  D_RV_Api_Obj.SetFieldBSTR(L_GuibosType, "buy_stp_status", L_status[rand()&7]);
  
  //; Status of the SELLERS transaction
  D_RV_Api_Obj.SetFieldBSTR(L_GuibosType, "sell_stp_status", L_status[rand()&7]);
}

#ifdef WIN32
#include <conio.h>
#endif
size_t L_get_line (char *line, int max_len)
{
  char *cur_pos = line;
  int             c_val;

  for (;;)
  {
#ifdef WIN32
    // this is just for wait for a while in windows
    Sleep (10);
#else
    usleep  (10000);
#endif
  
    if (!_kbhit ())
    {
      D_RV_Api_Obj.ProcessMyEvents ();      
      continue;
    }

    c_val = _getch ();

    if (c_val == '\n')
      break;
    if (c_val == '\r')
      break;
    if (c_val == 8)
    {
      if (cur_pos <= line)
        continue;

      printf ("\010 \010");
      --cur_pos;
      continue;
    }
    
#ifdef WIN32
    putchar (c_val);
#endif

    *cur_pos++ = c_val;
    if (--max_len <= 0)
      break;
  }
#ifdef WIN32
  putchar ('\n');
#endif
  *cur_pos = 0;
  return (size_t)(cur_pos - line);
}

#ifndef WIN32

static void init_keyboard()
{
  tcgetattr(0,&initial_settings);
  new_settings = initial_settings;
  new_settings.c_lflag &= ~ICANON;
  new_settings.c_lflag &= ~ISIG;
  new_settings.c_cc[VMIN] = 1;
  new_settings.c_cc[VTIME] = 0;
  tcsetattr(0, TCSANOW, &new_settings);
}

static void close_keyboard()
{
  tcsetattr(0, TCSANOW, &initial_settings);
}

static int _kbhit()
{
  char ch;
  int nread;

  if(peek_character != -1)
    return 1;

  nread = read(0,&ch,1);
  
  if(nread == 1) 
  {
    peek_character = ch;
    return 1;
  }

  return 0;
}

static int _getch()
{
  char ch;

  if(peek_character != -1) 
  {
    ch = peek_character;
    peek_character = -1;
    return ch;
  }

  read(0, &ch, 1);
  return ch;
}

#endif

///////////////////////////////////////////////////////////////////////////////////////////////
