﻿Public Class frmViewFields

    Public Sub New(ByVal m_blnWithDepth As Boolean, ByVal m_ColumnQty As Integer)
        InitializeComponent()

        'Set required variables
        Me.m_blnWithDepth = m_blnWithDepth
        Me.m_ColumnQty = m_ColumnQty
    End Sub

#Region "Module Variables and Structures"
    'Debug class designed so that it raises an event whenever its value is changed
    Public Class BooleanWithEvents
        Private BooleanValue As Boolean

        Property Value() As Boolean
            Get
                Return BooleanValue
            End Get

            Set(ByVal Value As Boolean)
                If Value <> BooleanValue Then
                    BooleanValue = Value
                    RaiseEvent ChangedValue()
                End If
            End Set
        End Property

        Public Sub New()
            BooleanValue = False
        End Sub

        Public Event ChangedValue()
    End Class

    'Class used to store additional information in the Tag property of a TreeNode
    Public Class NodeTag
        'Absolute index of a TreeNode when the TreeView is sorted
        Private PrivateSortedIndex As Integer
        'Absolute index of a TreeNode when the TreeView is unsorted
        Private PrivateUnsortedIndex As Integer
        'To mark the TreeNode as selected by the user
        Private PrivateIsSelected As Boolean

        Property SortedIndex() As Integer
            Get
                Return PrivateSortedIndex
            End Get

            Set(ByVal Value As Integer)
                PrivateSortedIndex = Value
            End Set
        End Property

        Property UnsortedIndex() As Integer
            Get
                Return PrivateUnsortedIndex
            End Get

            Set(ByVal Value As Integer)
                PrivateUnsortedIndex = Value
            End Set
        End Property

        Property IsSelected() As Boolean
            Get
                Return PrivateIsSelected
            End Get

            Set(ByVal Value As Boolean)
                PrivateIsSelected = Value
            End Set
        End Property

        Public Sub New()
            PrivateUnsortedIndex = -1
            PrivateSortedIndex = -1
            PrivateIsSelected = False
        End Sub

        Public Sub New(ByVal UnsortedIndex As Integer)
            PrivateUnsortedIndex = UnsortedIndex
            PrivateSortedIndex = -1
            PrivateIsSelected = False
        End Sub
    End Class

    'This class implements a sorter wich uses the NodeTag.UnsortedIndex property to sort the Tree
    Public Class TagSorter
        Implements System.Collections.IComparer

        Public Shared ReadOnly DefaultSorter As New TagSorter

        Public Function Compare(ByVal NodeA As Object, ByVal NodeB As Object) As Integer Implements System.Collections.IComparer.Compare
            'Cast input object to TreeNode objects
            Dim TreeNodeA As TreeNode = TryCast(NodeA, TreeNode)
            Dim TreeNodeB As TreeNode = TryCast(NodeB, TreeNode)

            'Perform comparison based on the UnsortedIndex property of the NodeTag
            Return CInt(CType(TreeNodeA.Tag, NodeTag).UnsortedIndex).CompareTo _
                        (CInt(CType(TreeNodeB.Tag, NodeTag).UnsortedIndex))

        End Function
    End Class

    'This array structure keeps an index of the currently selected node and a selection criteria
    Private Structure ArrayWithSelectedNodeIndex
        Public Element() As TreeNode
        'Index to currently selected node
        Public SelectedNode As Integer
        'Selection criteria
        Public Criteria As String
    End Structure

    'Value required in the constructor
    'Flag used to distinguish if the TreeView is filled with or without referenced fields
    Private m_blnWithDepth As Boolean

    'Value required in the constructor
    'This is the maximum number of records that can be dragged to Form1
    'to create additional columns
    'it is used in BuilColumnHeadersString to limit the resulting string
    Private m_ColumnQty As Integer

    'Array that stores nodes found by FindNode() function
    Private m_FoundNodesArray As ArrayWithSelectedNodeIndex = New ArrayWithSelectedNodeIndex

    'Note: a TreeView object does not support MultiSelect Mode;
    'therefore, this array is needed to keep track of all selected nodes.
    'This array stores all selected nodes. That is, all nodes found by FindNode() and also 
    'the nodes added manually. If there are no nodes found by FindNode() then this array will contain only 
    'manually selected nodes.
    Private m_SelectedNodesArray() As TreeNode = New TreeNode() {}

    'Boolean flag used to distinguish between manual (user driven)
    'and automatic (btnNext.Click driven) Tree.AfterSelect events
    'Used to prevent resetting of the m_SelectedNodesArray
    Private m_blnManualUserSelection As Boolean = True

    'Boolean flag used to prevent unhighlighting of a node in Tree.BeforeSelect events
    '(when btnNext button is clicked)
    Private m_blnNextButtonClicked As Boolean = False

    'Boolean flags used to aid in manual node selection 
    'They provide the feel of a MultiSelect mode TreeView object
    Private m_blnMouseIsDown As Boolean = False
    Private m_MouseButton As System.Windows.Forms.MouseButtons
    'This boolean flags have the capability of raising an event when the stored value is changed
    Friend WithEvents m_blnShiftIsDown As BooleanWithEvents = New BooleanWithEvents
    Friend WithEvents m_blnCtrlIsDown As BooleanWithEvents = New BooleanWithEvents

    'Boolean flag used to defer resetting of the font to m_Font for child nodes
    'The purpose of this flag was to allow the correct highlighting of a child node
    '(whithout chopping the last letters) but it isn't working
    Friend WithEvents m_blnFirstExpansion As BooleanWithEvents = New BooleanWithEvents

    'This boolean flag prevents the selection of a node when the +/- sign is clicked
    'in Tree.BeforeExpand and Tree.BeforeCollapse events
    Private m_blnTreeCollapseExpansionFlag As Boolean = False

    'Boolean flag used to prevent lock in txtFindstr control (which by default has the focus)
    'txtFindstr regains focus every time it looses focus unless this flag is False
    Private m_blnfrmViewFieldsActive As Boolean = True

    'Boolean flag to correct Tree.CollapseAll behaviour (sets Tree.SelectedNode to nothing)
    'used to prevent ClearNodeSelection run in the Tree.AfterSelect event
    Private m_blnCollapseAllClicked As Boolean = False

    'Array that stores relative position of control objects within the form
    'These aids the repositioning of these elements when the form is resized
    Private Const ME_TREE As Integer = 0
    Private Const TXT_FINDSTR As Integer = 1
    'BTN_FIND and BTN_NEXT are also used as constants for m_btnFindState variable
    'BTN_NEXT is a virtual button and it is BTN_FIND with a different m_btnFindState state
    Private Const BTN_FIND As Integer = 2
    Private Const BTN_NEXT As Integer = 8
    Private Const BTN_EXPANDALL As Integer = 3
    Private Const BTN_COLLAPSEALL As Integer = 4
    'BTN_SORT and BTN_UNSORT are also used as constants for m_btnSortState variable
    'BTN_UNSORT is a virtual button and it is BTN_SORT with a different m_btnSortState state
    Private Const BTN_SORT As Integer = 5
    Private Const BTN_UNSORT As Integer = 9
    Private Const BTN_SAVE As Integer = 6
    Private Const TXT_MESSAGE As Integer = 7
    Private m_ControlVerticalPosition() As Integer = New Integer(7) {}

    'Used to determine in which state (of a total of two states) is the button btnFind
    'The two possible values are: BTN_FIND and BTN_NEXT
    Private m_btnFindState As Integer = BTN_FIND

    'Used to determine in which state (of a total of two states) is the button btnSort
    'The two possible values are: BTN_SORT and BTN_UNSORT
    Private m_btnSortState As Integer = BTN_SORT

    'Standard and highlighted node fonts (default Tree.Font is m_BoldFont)
    Private m_Font As Font = New Font("Microsoft Sans Serif", 8, FontStyle.Regular)
    Private m_BoldFont As Font = New Font("Microsoft Sans Serif", 8, FontStyle.Bold)

    'Module variable that holds a count of nodes in the tree
    'It will be used as an index stored in the Tag property of a TreeNode
    Private m_NodeCount As Integer = 0

    'Default TreeViewNodeSorter
    Private m_TreeSorter As System.Collections.IComparer
#End Region

#Region "Module Subroutines"
    Public Sub InsertNode(ByRef ParentNode As TreeNode, ByVal strFieldName As String, ByVal strFieldValue As String, ByVal blnWithDepth As Boolean)
        If ParentNode Is Nothing Then
            If strFieldName.IndexOf(".") = -1 Then
                'If ParentNode is Nothing and FieldName does not contain referenced FieldNames then insert node at the root level
                'Create a New node so that the bold font can be assigned to it
                'instead of using the Me.Tree.Nodes.Add(txt) method which would be cleaner to read
                Dim NewNode As TreeNode = New TreeNode(strFieldName & " : " & strFieldValue)
                NewNode.NodeFont = m_Font
                NewNode.Tag = New NodeTag(m_NodeCount)
                Me.Tree.Nodes.Add(NewNode)

                m_NodeCount = m_NodeCount + 1
            Else
                'If ParentNode is Nothing and FieldName contains referenced FieldNames (by convention separated by dots) then
                'get root level ParentNode (which is the last inserted root level node), strip root level ParentNode text from strFieldName
                'and recursively call InsertNode until the proper level is reached
                'Only traverse depth if blnWithDepth flag is set
                If blnWithDepth Then InsertNode(Me.Tree.Nodes(Me.Tree.Nodes.Count - 1), Mid(strFieldName, strFieldName.IndexOf(".") + 2), strFieldValue, blnWithDepth)
            End If
        Else
            If strFieldName.IndexOf(".") = -1 Then
                'If ParentNode exists and there are no more referenced FieldNames then insert node at the ParentNode level
                Dim NewNode As TreeNode = New TreeNode(strFieldName & " : " & strFieldValue)
                NewNode.NodeFont = m_BoldFont
                NewNode.Tag = New NodeTag(m_NodeCount)
                ParentNode.Nodes.Add(NewNode)

                m_NodeCount = m_NodeCount + 1
            Else
                'If ParentNode exists and there are more referenced FieldNames (by convention separated by dots) then
                'recursively call InsertNode until the proper level is reached
                InsertNode(ParentNode.Nodes(ParentNode.Nodes.Count - 1), Mid(strFieldName, strFieldName.IndexOf(".") + 2), strFieldValue, blnWithDepth)
            End If
        End If
    End Sub
    Private Sub ClearNodeSelection()
        'Clear m_FoundNodesArray.Elements
        If UBound(m_FoundNodesArray.Element) > -1 Then
            ReDim m_FoundNodesArray.Element(-1)
        End If
        'index to currently selected node
        m_FoundNodesArray.SelectedNode = -1
        'and search criteria
        m_FoundNodesArray.Criteria = ""

        'Clear m_SelectedNodesArray
        If UBound(m_SelectedNodesArray) > -1 Then

            'Unhiglight all nodes in m_SelectedNodesArray
            For Each Node As TreeNode In m_SelectedNodesArray
                'Unhiglight node
                UnHighlightNode(Node)
            Next

            ReDim m_SelectedNodesArray(-1)
        End If

        'Clear txtMessage
        txtMessage.Text = ""
    End Sub
    Private Sub HighlightNode(ByRef Node As TreeNode)
        'Check input parameter
        If Node Is Nothing Then Return

        'Highlight node
        Node.BackColor = System.Drawing.SystemColors.Highlight
        Node.ForeColor = System.Drawing.SystemColors.HighlightText
        Node.NodeFont = m_BoldFont

        'Set Tag property which will be used to discriminate between selected and unselected nodes
        CType(Node.Tag, NodeTag).IsSelected = True
    End Sub
    Private Sub UnHighlightNode(ByRef Node As TreeNode)
        'Check input parameter
        If Node Is Nothing Then Return

        'Unhighlight node
        Node.BackColor = System.Drawing.SystemColors.Window
        Node.ForeColor = System.Drawing.SystemColors.WindowText
        Node.NodeFont = m_Font

        'Reset Tag property
        CType(Node.Tag, NodeTag).IsSelected = False
    End Sub
    Private Sub UnSelectNode(ByRef NodeToDelete As TreeNode)
        'Check input parameters
        If NodeToDelete Is Nothing Or UBound(m_SelectedNodesArray) < 0 Then Return

        'Flag used to signal that NodeToDelete was found at the ith position
        Dim blnBeginMovingElements As Boolean = False

        'If NodeToDeleteMove is found begin moving elements to the left
        For i As Integer = 0 To UBound(m_SelectedNodesArray)
            If blnBeginMovingElements Then
                m_SelectedNodesArray(i - 1) = m_SelectedNodesArray(i)
            Else
                If CompareTreeNodes(m_SelectedNodesArray(i), NodeToDelete) = 0 Then blnBeginMovingElements = True
            End If
        Next

        'if blnBeginMovingElements flag was not raised then NodeToDelete is not in m_SelectedNodesArray
        If Not blnBeginMovingElements Then Return

        'Move last element to the left and redimension m_SelectedNodesArray
        ReDim Preserve m_SelectedNodesArray(UBound(m_SelectedNodesArray) - 1)

        'Unhighlight NodeToDelete
        UnHighlightNode(NodeToDelete)
    End Sub
    Private Sub SetTreeNodesFont(ByRef SetFont As Font, Optional ByRef StartNode As TreeNode = Nothing)
        'If Nothing is supplied then start from the beginning
        If StartNode Is Nothing Then StartNode = Tree.Nodes(0)

        'Loop through each node and change font
        Do
            StartNode.NodeFont = SetFont
            If StartNode.FirstNode IsNot Nothing Then SetTreeNodesFont(SetFont, StartNode.FirstNode)
            StartNode = StartNode.NextNode
        Loop Until StartNode Is Nothing
    End Sub
    Private Sub InitializeTagPropertyOld(Optional ByRef StartNode As TreeNode = Nothing)
        'Not used now since the initialization is done when the new node is inserted
        'Check the first node in the tree
        If StartNode Is Nothing Then
            StartNode = Tree.Nodes(0)
            If StartNode.Tag Is Nothing Then
                StartNode.Tag = New NodeTag(0)
                m_NodeCount = 1
            End If
            If StartNode.FirstNode IsNot Nothing Then InitializeTagPropertyOld(StartNode.FirstNode)
            StartNode = StartNode.NextNode
        End If

        'Then continue with the rest of the tree
        Do
            StartNode.Tag = New NodeTag(m_NodeCount)
            m_NodeCount = m_NodeCount + 1
            If StartNode.FirstNode IsNot Nothing Then InitializeTagPropertyOld(StartNode.FirstNode)
            StartNode = StartNode.NextNode
        Loop Until StartNode Is Nothing
    End Sub
    Private Sub SetSortedIndex(Optional ByRef StartNode As TreeNode = Nothing)
        'Set initial node and reset m_NodeCount counter
        If StartNode Is Nothing Then
            StartNode = Tree.Nodes(0)
            m_NodeCount = 0
        End If

        'Then continue traversing the tree
        Do
            CType(StartNode.Tag, NodeTag).SortedIndex = m_NodeCount
            m_NodeCount = m_NodeCount + 1
            If StartNode.FirstNode IsNot Nothing Then SetSortedIndex(StartNode.FirstNode)
            StartNode = StartNode.NextNode
        Loop Until StartNode Is Nothing
    End Sub
    Private Function FindNode(ByRef mytree As TreeView, ByVal strToSearch As String) As Integer
        'If strToSearch is empty do nothing
        If strToSearch.Length <= 0 Then Return 0

        'Reset m_FoundNodesArray
        ClearNodeSelection()

        'Remember search criteria
        m_FoundNodesArray.Criteria = strToSearch.ToLower

        'Select first element in Tree
        Dim TempNode As TreeNode = Tree.Nodes(0)
        'Search entire visible nodes 
        Do Until TempNode Is Nothing
            If TempNode.Text.IndexOf(strToSearch.ToLower) >= 0 Then
                'Fill m_FoundNodesArray
                ReDim Preserve m_FoundNodesArray.Element(UBound(m_FoundNodesArray.Element) + 1)
                m_FoundNodesArray.Element(UBound(m_FoundNodesArray.Element)) = TempNode

                'Fill m_SelectedNodesArray
                ReDim Preserve m_SelectedNodesArray(UBound(m_SelectedNodesArray) + 1)
                m_SelectedNodesArray(UBound(m_SelectedNodesArray)) = TempNode

            End If
            TempNode = TempNode.NextVisibleNode
        Loop

        'Return count of nodes found
        Return UBound(m_FoundNodesArray.Element) + 1
    End Function
    Private Function GetText(ByVal CurrentNode As TreeNode) As String
        Dim ParentNode As TreeNode = CurrentNode.Parent
        'Strip CurrentNode.Text into FieldName,FieldValue for csv file
        Dim TempText As String = Mid(CurrentNode.Text, 1, CurrentNode.Text.IndexOf(":") - 1) & "," & Mid(CurrentNode.Text, CurrentNode.Text.IndexOf(":") + 3)

        If ParentNode Is Nothing Then
            'If ParentNode is Nothing GetText is finished
            Return TempText
        Else
            'Else get all parents 
            Return GetParentsText(ParentNode) & "." & TempText
        End If
    End Function
    Private Function GetParentsText(ByRef CurrentNode As TreeNode) As String
        'If CurrentNode is Nothing GetParents is finished
        If CurrentNode Is Nothing Then Return ""

        'Recursively call GetParents until CurrentNode is Nothing
        Dim TempText As String = GetParentsText(CurrentNode.Parent)

        'Depending on ParentFieldName lenght
        If TempText.Length <= 0 Then
            'Return FieldName (stripped from CurrentNode.Text)
            Return Mid(CurrentNode.Text, 1, CurrentNode.Text.IndexOf(":") - 1)
        Else
            'Or return ParentFieldName.FieldName
            Return TempText & "." & Mid(CurrentNode.Text, 1, CurrentNode.Text.IndexOf(":") - 1)
        End If
    End Function
    Private Function CompareTreeNodes(ByRef NodeA As TreeNode, ByRef NodeB As TreeNode) As Integer
        'This function is required to compare Nodes that are at different depht levels
        'It returns 1 if NodeA.Index > NodeB.Index
        'zero if NodeA.Index = NodeB.Index
        '-1 if NodeA.Index < NodeB.Index

        'If either one or both of the input nodes is Nothing then
        If NodeA IsNot Nothing And NodeB Is Nothing Then
            Return 1
        ElseIf NodeA Is Nothing And NodeB IsNot Nothing Then
            Return -1
        ElseIf NodeA Is Nothing And NodeB Is Nothing Then
            Return 0
        End If

        Dim Result As Integer = 0

        'This more efficient way improves speed
        'Compare using the proper index depending on the state of btnSort button
        'which in turn tells the sort status of the TreeView TreeNodes (sorted or unsorted)
        If m_btnSortState = BTN_SORT Then
            Result = CType(NodeA.Tag, NodeTag).UnsortedIndex - CType(NodeB.Tag, NodeTag).UnsortedIndex
        Else 'If m_btnSortState = BTN_UNSORT Then
            Result = CType(NodeA.Tag, NodeTag).SortedIndex - CType(NodeB.Tag, NodeTag).SortedIndex
        End If

        'If both are equal return Result
        If Result = 0 Then Return Result

        'Else return 1 or -1
        Return Result / System.Math.Abs(Result)
    End Function
    Private Function CompareTreeNodesOld(ByRef NodeA As TreeNode, ByRef NodeB As TreeNode) As Integer
        'This function is required to compare Nodes that are at different depht levels
        'It returns 1 if NodeA.Index > NodeB.Index
        'zero if NodeA.Index = NodeB.Index
        '-1 if NodeA.Index < NodeB.Index

        'Older way to check if node is already selected
        Dim i As Integer
        Dim Comparison As Integer

        'If either one or both of the input nodes is Nothing then
        If NodeA IsNot Nothing And NodeB Is Nothing Then
            Return 1
        ElseIf NodeA Is Nothing And NodeB IsNot Nothing Then
            Return -1
        ElseIf NodeA Is Nothing And NodeB Is Nothing Then
            Return 0
        End If

        'For each node create an IndexArray which will contain a list of 
        'its parents's indexes and lastly its own index
        'If the node has no parents then the array will contain only one element
        Dim NodeAIndexArray() As Integer = New Integer() {}
        Dim NodeBIndexArray() As Integer = New Integer() {}

        'Get depth leves
        Dim NodeADepthLevel As Integer = GetTreeNodeDepthLevel(NodeA)
        Dim NodeBDepthLevel As Integer = GetTreeNodeDepthLevel(NodeB)
        Dim MaxDepthLevel As Integer = System.Math.Max(NodeADepthLevel, NodeBDepthLevel)
        Dim MinDepthLevel As Integer = System.Math.Min(NodeADepthLevel, NodeBDepthLevel)

        'Redimension arrays according to necessity
        ReDim Preserve NodeAIndexArray(NodeADepthLevel)
        ReDim Preserve NodeBIndexArray(NodeBDepthLevel)

        'For Nodes at different depth leves fill IndexArrays and compare 
        If MaxDepthLevel > 0 Then
            For i = 0 To MinDepthLevel
                NodeAIndexArray(i) = GetTreeNodeParent(NodeA, NodeADepthLevel - i).Index
                NodeBIndexArray(i) = GetTreeNodeParent(NodeB, NodeBDepthLevel - i).Index
                Comparison = CompareTreeNodeIndexArrays(NodeAIndexArray, NodeBIndexArray, i)
                'If Index at the ith level is different then there is no need to keep comparing
                If Comparison <> 0 Then Return Comparison
            Next
        End If

        'If NodeA's parents and NodeB's parents are all equal compare DepthLevel
        If NodeADepthLevel > NodeBDepthLevel Then
            Return 1
        ElseIf NodeADepthLevel < NodeBDepthLevel Then
            Return -1
        End If

        'Finally compare NodeA.Index and NodeB.Index
        If NodeA.Index > NodeB.Index Then
            Return 1
        ElseIf NodeA.Index < NodeB.Index Then
            Return -1
        End If

        'The elements are equal
        Return 0

    End Function
    Private Function CompareTreeNodeIndexArrays(ByRef IndexArrayA As Integer(), ByRef IndexArrayB As Integer(), ByVal Index As Integer) As Integer
        'This function compares IndexArray at the indexth level
        'It returns 1 if IndexArrayA() > IndexArrayB()
        'zero if IndexArrayA() = IndexArrayB()
        '-1 if IndexArrayA() < IndexArrayB()

        Try
            'Trivial comparison
            If IndexArrayA(Index) > IndexArrayB(Index) Then
                Return 1
            ElseIf IndexArrayA(Index) < IndexArrayB(Index) Then
                Return -1
            End If

            Return 0
        Catch ex As Exception
            Debug.Print("CompareIndexArrays: Index is out of range or at least one of the IndexArray elements is Nothing")
        End Try

    End Function
    Private Function GetTreeNodeDepthLevel(ByRef ThisNode As TreeNode) As Integer
        'Check that input parameter is correct
        If ThisNode Is Nothing Then Return -1

        Dim Depth As Integer = -1
        Dim Parent As TreeNode = ThisNode

        'Search for a Parent until Nothing is found
        Do
            Depth = Depth + 1
            Parent = Parent.Parent
        Loop While Parent IsNot Nothing

        Return Depth
    End Function
    Private Function GetTreeNodeParent(ByRef ThisNode As TreeNode, Optional ByVal UpperLevel As Integer = 1) As TreeNode
        'This function is required to get the Parent of a TreeNode at the UpperLevel level
        'It returns the Parent of a TreeNode when optional AtDepth variable is not supplied

        'There are no Parents at negative UpperLevel values
        If UpperLevel < 0 Then Return Nothing

        'GetTreeNodeParent at the UpperLevel 0 is the current node
        If UpperLevel = 0 Then Return ThisNode

        Dim Parent As TreeNode = ThisNode

        'Search for a Parent until UpperLevel is reached or Nothing is found
        For i As Integer = 1 To UpperLevel
            If Parent IsNot Nothing Then
                Parent = Parent.Parent
            Else
                'Stop searching
                'There are no more Parents beyond this level
                Exit For
            End If
        Next

        Return Parent

    End Function
    Private Function BuildColumnHeadersString() As String
        'If there are no members in m_SelectedNodesArray then nothing to be done
        If UBound(m_SelectedNodesArray) < 0 Then Return ""

        Dim strResult As String = ""
        Dim strTemp As String = ""

        'Set UpperLimit to the size of m_SelectedNodesArray or to MAX_RECORDS_QTY, whichever is less
        Dim UpperLimit As Integer = System.Math.Min(m_ColumnQty, UBound(m_SelectedNodesArray))

        'At least one element will be processed if m_ColumnQty is zero(from 0 to 0)
        For i As Integer = 0 To UpperLimit
            strTemp = GetText(m_SelectedNodesArray(i))
            strResult = strResult & "," & Mid(strTemp, 1, strTemp.IndexOf(","))
        Next
        'Strip first commna
        strResult = Mid(strResult, 2)

        Return strResult
    End Function
    Private Function NodeIsAlreadySelected(ByRef NodeToCheck As TreeNode) As Boolean
        'Check input parameters
        If NodeToCheck Is Nothing Then Return False

        'This more efficient way improves speed
        Return CType(NodeToCheck.Tag, NodeTag).IsSelected

        'Older way to check if node is already selected
        For Each Node As TreeNode In m_SelectedNodesArray
            If CompareTreeNodes(Node, NodeToCheck) = 0 Then
                Return True
            End If
        Next
        'If NodeToCheck is not in m_SelectedNodesArray
        Return False
    End Function
    Private Function GetNodeVisiblePosition(ByRef Node As TreeNode) As Integer
        'Returns the relative position of a visible node in the current view
        'it returns 0 if the node is not visible or the current line in which this node is displayed

        If Node Is Nothing Then Return 0

        'if Node is not Nothing but it is not visible return as it was in the middle of the screen
        If Not Node.IsVisible Then Return Tree.VisibleCount / 2

        Dim TopNode As TreeNode = Tree.TopNode
        Dim CurrentNode As TreeNode = Tree.TopNode

        'Traverse from TopNode downwards until Node is reached
        For i As Integer = 1 To Tree.VisibleCount
            If CompareTreeNodes(CurrentNode, Node) = 0 Then Return i
            CurrentNode = CurrentNode.NextVisibleNode
        Next

        'If nothing has satisfied the conditions then set return value to this
        Return Tree.VisibleCount
    End Function
#End Region

#Region "Module Event Handlers"
    'frmViewFields form events
    Private Sub frmViewFields_Load(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles MyBase.Load
        'Module variable initialization
        m_FoundNodesArray.Element = New TreeNode() {}
        m_FoundNodesArray.SelectedNode = -1
        m_FoundNodesArray.Criteria = ""

        'Store relative position of frmViewFields control objects
        m_ControlVerticalPosition(ME_TREE) = Me.Size.Height - Tree.Height
        m_ControlVerticalPosition(TXT_FINDSTR) = Me.Size.Height - txtFindstr.Location.Y
        m_ControlVerticalPosition(BTN_FIND) = Me.Size.Height - btnFind.Location.Y
        m_ControlVerticalPosition(BTN_EXPANDALL) = Me.Size.Height - btnExpandAll.Location.Y
        m_ControlVerticalPosition(BTN_COLLAPSEALL) = Me.Size.Height - btnCollapseAll.Location.Y
        m_ControlVerticalPosition(BTN_SORT) = Me.Size.Height - btnSort.Location.Y
        m_ControlVerticalPosition(BTN_SAVE) = Me.Size.Height - btnSave.Location.Y
        m_ControlVerticalPosition(TXT_MESSAGE) = Me.Size.Height - txtMessage.Location.Y

        'Set form maximum size
        Me.MaximumSize = New Size(Me.Size.Width, Screen.GetWorkingArea(Me.Location).Height + 8)

        m_blnFirstExpansion.Value = True

        'If data received with m_blnWithDepth then enable controls as needed
        If m_blnWithDepth Then
            btnExpandAll.Enabled = True
        Else
            m_blnFirstExpansion.Value = False
        End If

        'Save default sorter
        m_TreeSorter = Me.Tree.TreeViewNodeSorter
    End Sub
    Private Sub frmViewFields_Shown(ByVal sender As Object, ByVal e As System.EventArgs) Handles MyBase.Shown
        'The tree was created with bold font so that the Text property is highligted correctly when using bold font
        'Prior to this change the Text property was chopped and did not display the entire text when highlighted
        'This procedure execution was deferred to improve speed when loading frmViewFields for the first time
        'SetTreeNodesFont(m_Font)

        'By default frmViewFields selects the first element in the Tree, but first initialize TreeNode.Tag
        Tree.Nodes(0).Tag = New NodeTag(m_NodeCount)
        m_NodeCount = m_NodeCount + 1
        Tree.SelectedNode = Tree.Nodes(0)
        HighlightNode(Tree.SelectedNode)

        'Fill SelectedNodesArray with current selection
        ReDim m_SelectedNodesArray(0)
        m_SelectedNodesArray(0) = Tree.SelectedNode
    End Sub
    Private Sub frmViewFields_SizeChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles MyBase.SizeChanged
        'If form is not visible there is no need to reposition controls
        If Not Me.Visible Then Return

        'Reposition frmViewFields control objects 
        Tree.Height = Me.Height - m_ControlVerticalPosition(ME_TREE)
        txtFindstr.Location = New Point(txtFindstr.Location.X, Me.Size.Height - m_ControlVerticalPosition(TXT_FINDSTR))
        btnFind.Location = New Point(btnFind.Location.X, Me.Size.Height - m_ControlVerticalPosition(BTN_FIND))
        btnExpandAll.Location = New Size(btnExpandAll.Location.X, Me.Size.Height - m_ControlVerticalPosition(BTN_EXPANDALL))
        btnCollapseAll.Location = New Size(btnCollapseAll.Location.X, Me.Size.Height - m_ControlVerticalPosition(BTN_COLLAPSEALL))
        btnSort.Location = New Size(btnSort.Location.X, Me.Size.Height - m_ControlVerticalPosition(BTN_SORT))
        btnSave.Location = New Size(btnSave.Location.X, Me.Size.Height - m_ControlVerticalPosition(BTN_SAVE))
        txtMessage.Location = New Size(txtMessage.Location.X, Me.Size.Height - m_ControlVerticalPosition(TXT_MESSAGE))
    End Sub
    Private Sub frmViewFields_Activated(ByVal sender As Object, ByVal e As System.EventArgs) Handles MyBase.Activated
        m_blnfrmViewFieldsActive = True
        txtFindstr.Focus()
    End Sub
    Private Sub frmViewFields_Deactivate(ByVal sender As Object, ByVal e As System.EventArgs) Handles MyBase.Deactivate
        m_blnfrmViewFieldsActive = False
    End Sub
    Private Sub frmViewFields_KeyDown(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles MyBase.KeyDown

        'Set mutually exclusive flags for m_blnCtrlIsDown and m_blnShiftIsDown
        If e.Shift And Not m_blnCtrlIsDown.Value Then
            m_blnShiftIsDown.Value = True
        End If
        If e.Control And Not m_blnShiftIsDown.Value Then
            m_blnCtrlIsDown.Value = True
        End If

        'Up navigation keys
        If e.KeyCode = System.Windows.Forms.Keys.Home Then
            'Reset txtFindstr
            txtFindstr.Text = ""

            'Select the first element in the Tree
            Tree.SelectedNode = Tree.Nodes(0)
            HighlightNode(Tree.SelectedNode)

            'Fill SelectedNodesArray with current selection
            ReDim m_SelectedNodesArray(0)
            m_SelectedNodesArray(0) = Tree.SelectedNode

            'Signal that the event has processed everything
            e.SuppressKeyPress = True
            e.Handled = True
        End If
        If e.KeyCode = System.Windows.Forms.Keys.Up Then
            'Signal that the event has processed everything
            e.SuppressKeyPress = True
            e.Handled = True

            'If no node is selected then do nothing
            If Tree.SelectedNode Is Nothing Then Return

            'Proceed only if there is a PrevVisibleNode
            If Tree.SelectedNode.PrevVisibleNode IsNot Nothing Then
                'Update selected node
                Tree.SelectedNode = Tree.SelectedNode.PrevVisibleNode

                'If Shift is not pressed then select only PrevVisibleNode
                If Not e.Modifiers = Keys.Shift Then
                    'Highlight and update IsSelected flag
                    HighlightNode(Tree.SelectedNode)

                    'Update node selection
                    ReDim m_SelectedNodesArray(0)
                    m_SelectedNodesArray(0) = Tree.SelectedNode
                Else
                    'Include this node in m_SelectedNodesArray unless it is already there
                    If Not NodeIsAlreadySelected(Tree.SelectedNode) Then
                        'Highlight and update IsSelected flag
                        HighlightNode(Tree.SelectedNode)

                        ReDim Preserve m_SelectedNodesArray(UBound(m_SelectedNodesArray) + 1)
                        m_SelectedNodesArray(UBound(m_SelectedNodesArray)) = Tree.SelectedNode
                    End If
                End If

                'Reset txtFindstr string
                txtFindstr.Text = ""

                'Update txtMessage
                If UBound(m_SelectedNodesArray) > 0 Then
                    txtMessage.Text = "Selected " & UBound(m_SelectedNodesArray) + 1 & " records."
                Else
                    txtMessage.Text = ""
                End If
            End If
        End If
        If e.KeyData = System.Windows.Forms.Keys.PageUp Then
            'If no node is selected then move up one page based on Tree.TopNode
            If Tree.SelectedNode Is Nothing Then
                Dim Node As TreeNode = Tree.TopNode
                Dim i As Integer = 1

                Do While Node.PrevVisibleNode IsNot Nothing And i < Tree.VisibleCount
                    Node = Node.PrevVisibleNode
                    i = i + 1
                Loop
                Tree.TopNode = Node
            ElseIf Tree.SelectedNode.PrevVisibleNode IsNot Nothing Then
                Dim Node As TreeNode = Tree.SelectedNode
                Dim i As Integer = 1

                'Clear search criteria
                txtFindstr.Text = ""

                Do While Node.PrevVisibleNode IsNot Nothing And i < Tree.VisibleCount
                    Node = Node.PrevVisibleNode
                    i = i + 1
                Loop

                'Destination reached
                Tree.SelectedNode = Node
                HighlightNode(Tree.SelectedNode)
            End If

            'Signal that the event has processed everything
            e.SuppressKeyPress = True
            e.Handled = True
        End If

        'Down navigation keys
        If e.KeyCode = System.Windows.Forms.Keys.End Then
            'Reset txtFindstr
            txtFindstr.Text = ""

            'Select the last element in the Tree
            If Not Tree.Nodes(Tree.Nodes.Count - 1).IsExpanded Then
                Tree.SelectedNode = Tree.Nodes(Tree.Nodes.Count - 1)
            Else
                Dim CurrentNode As TreeNode = Tree.Nodes(Tree.Nodes.Count - 1)
                Do
                    CurrentNode = CurrentNode.NextVisibleNode
                Loop While CurrentNode.NextVisibleNode IsNot Nothing
                Tree.SelectedNode = CurrentNode
            End If
            HighlightNode(Tree.SelectedNode)

            'Fill SelectedNodesArray with current selection
            ReDim m_SelectedNodesArray(0)
            m_SelectedNodesArray(0) = Tree.SelectedNode

            'Signal that the event has processed everything
            e.SuppressKeyPress = True
            e.Handled = True
        End If
            If e.KeyCode = System.Windows.Forms.Keys.Down Then
                'Signal that the event has processed everything
                e.SuppressKeyPress = True
                e.Handled = True

                'If no node is selected then do nothing
                If Tree.SelectedNode Is Nothing Then Return

                'Proceed only if there is a NextVisibleNode
                If Tree.SelectedNode.NextVisibleNode IsNot Nothing Then
                    'Update selected node
                    Tree.SelectedNode = Tree.SelectedNode.NextVisibleNode

                    'If Shift is not pressed then select only NextVisibleNode
                    If Not e.Modifiers = Keys.Shift Then
                        'Highlight and update IsSelected flag
                        HighlightNode(Tree.SelectedNode)

                        'Update node selection
                        ReDim m_SelectedNodesArray(0)
                        m_SelectedNodesArray(0) = Tree.SelectedNode
                    Else
                        'Include this node in m_SelectedNodesArray unless it is already there
                        If Not NodeIsAlreadySelected(Tree.SelectedNode) Then
                            'Highlight and update IsSelected flag
                            HighlightNode(Tree.SelectedNode)

                            ReDim Preserve m_SelectedNodesArray(UBound(m_SelectedNodesArray) + 1)
                            m_SelectedNodesArray(UBound(m_SelectedNodesArray)) = Tree.SelectedNode
                        End If
                    End If

                    'Reset txtFindstr string
                    txtFindstr.Text = ""

                    'Update txtMessage
                    If UBound(m_SelectedNodesArray) > 0 Then
                        txtMessage.Text = "Selected " & UBound(m_SelectedNodesArray) + 1 & " records."
                    Else
                        txtMessage.Text = ""
                    End If
                End If
            End If
            If e.KeyData = System.Windows.Forms.Keys.PageDown Then
                'If no node is selected then move down one page based on Tree.TopNode
                If Tree.SelectedNode Is Nothing Then
                    Dim Node As TreeNode = Tree.TopNode
                    Dim i As Integer = 1

                    Do While Node.NextVisibleNode IsNot Nothing And i < Tree.VisibleCount
                        Node = Node.NextVisibleNode
                        i = i + 1
                    Loop
                    Tree.TopNode = Node
                ElseIf Tree.SelectedNode.NextVisibleNode IsNot Nothing Then
                    Dim Node As TreeNode = Tree.SelectedNode
                    Dim i As Integer = 1

                    'Clear search criteria
                    txtFindstr.Text = ""

                    Do While Node.NextVisibleNode IsNot Nothing And i < Tree.VisibleCount
                        Node = Node.NextVisibleNode
                        i = i + 1
                    Loop

                    'Destination reached
                    Tree.SelectedNode = Node
                    HighlightNode(Tree.SelectedNode)
                End If

                'Signal that the event has processed everything
                e.SuppressKeyPress = True
                e.Handled = True
            End If

            'Left and right navigation keys
            If e.KeyData = System.Windows.Forms.Keys.Left Then
                'Handle only if there is a Tree.SelectedNode
                If Tree.SelectedNode Is Nothing Then Return

                If Tree.SelectedNode.GetNodeCount(False) > 0 And Tree.SelectedNode.IsExpanded Then
                    Tree.SelectedNode.Collapse()
                    e.SuppressKeyPress = True
                    e.Handled = True
                End If
            End If
            If e.KeyData = System.Windows.Forms.Keys.Right Then
                'Handle only if there is a Tree.SelectedNode
                If Tree.SelectedNode Is Nothing Then Return

                If Tree.SelectedNode.GetNodeCount(False) > 0 And Not Tree.SelectedNode.IsExpanded Then
                    Tree.SelectedNode.Expand()
                    e.SuppressKeyPress = True
                    e.Handled = True
                End If
            End If

            'Delete key
            If e.KeyData = System.Windows.Forms.Keys.Delete Then
                txtFindstr.Text = ""
                e.SuppressKeyPress = True
                e.Handled = True
            End If

            'Enter key
            If e.KeyData = System.Windows.Forms.Keys.Enter Then
                btnFind_Click(sender, New System.EventArgs)
                e.SuppressKeyPress = True
                e.Handled = True
            End If

            'Find nodes with txtFind criteria, Alt+F (Left Alt or Right Alt)
            If (e.Modifiers = Keys.Alt Or (e.Modifiers = Keys.Alt + Keys.Control)) And e.KeyCode.ToString.ToUpper = "F" Then
                If m_btnFindState = BTN_FIND Then btnFind_Click(sender, New System.EventArgs)
                e.SuppressKeyPress = True
                e.Handled = True
            End If

            'Go to next found node, Alt+N (Left Alt or Right Alt)
            If (e.Modifiers = Keys.Alt Or (e.Modifiers = Keys.Alt + Keys.Control)) And e.KeyCode.ToString.ToUpper = "N" Then
                If m_btnFindState = BTN_NEXT Then btnFind_Click(sender, New System.EventArgs)
                e.SuppressKeyPress = True
                e.Handled = True
            End If

            'Expand tree, Alt+E (Left Alt or Right Alt)
            If (e.Modifiers = Keys.Alt Or (e.Modifiers = Keys.Alt + Keys.Control)) And e.KeyCode.ToString.ToUpper = "E" Then
                If btnExpandAll.Enabled = True Then btnExpandAll_Click(sender, New System.EventArgs)
                e.SuppressKeyPress = True
                e.Handled = True
            End If

            'Collapse tree, Alt+C (Left Alt or Right Alt)
            If (e.Modifiers = Keys.Alt Or (e.Modifiers = Keys.Alt + Keys.Control)) And e.KeyCode.ToString.ToUpper = "C" Then
                If btnCollapseAll.Enabled = True Then btnCollapseAll_Click(sender, New System.EventArgs)
                e.SuppressKeyPress = True
                e.Handled = True
            End If

            'Sort tree, Alt+T (Left Alt or Right Alt)
            If (e.Modifiers = Keys.Alt Or (e.Modifiers = Keys.Alt + Keys.Control)) And e.KeyCode.ToString.ToUpper = "T" Then
                btnSort_Click(sender, New System.EventArgs)
                e.SuppressKeyPress = True
                e.Handled = True
            End If

            'Save tree, Alt+S (Left Alt or Right Alt)
            If (e.Modifiers = Keys.Alt Or (e.Modifiers = Keys.Alt + Keys.Control)) And e.KeyCode.ToString.ToUpper = "S" Then
                btnSave_Click(sender, New System.EventArgs)
                e.SuppressKeyPress = True
                e.Handled = True
            End If

            'Copy selected nodes, Ctrl+C (Ctrl+V not implemented on the other side at the moment)
            If (e.Modifiers = Keys.Control) And e.KeyCode.ToString.ToUpper = "C" Then
                Debug.Print(BuildColumnHeadersString())
                Tree.DoDragDrop(BuildColumnHeadersString(), DragDropEffects.Copy)
                e.SuppressKeyPress = True
                e.Handled = True
            End If

    End Sub
    Private Sub frmViewFields_KeyUp(ByVal sender As Object, ByVal e As System.Windows.Forms.KeyEventArgs) Handles MyBase.KeyUp
        'Keyboard has been released change boolean flags accordingly
        If e.KeyData And Keys.ShiftKey Then m_blnShiftIsDown.Value = False
        If e.KeyData And Keys.ControlKey Then m_blnCtrlIsDown.Value = False
    End Sub
    Private Sub frmViewFields_MouseWheel(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Me.MouseWheel
        Dim Node As TreeNode = Tree.TopNode

        If e.Delta > 0 Then
            If Node.PrevVisibleNode IsNot Nothing Then Tree.TopNode = Node.PrevVisibleNode
        Else
            If Node.NextVisibleNode IsNot Nothing Then Tree.TopNode = Node.NextVisibleNode
        End If
    End Sub
    'Debug boolean variables
    Private Sub m_blnCtrlIsDown_ChangedValue() Handles m_blnCtrlIsDown.ChangedValue
        'Debug.Print("m_blnCtrlIsDown " & m_blnCtrlIsDown.Value)
    End Sub
    Private Sub m_blnShiftIsDown_ChangedValue() Handles m_blnShiftIsDown.ChangedValue
        'Debug.Print("m_blnShiftIsDown " & m_blnShiftIsDown.Value)
    End Sub
    Private Sub m_blnFirstExpansion_ChangedValue() Handles m_blnFirstExpansion.ChangedValue
        Debug.Print("m_blnFirstExpansion " & m_blnFirstExpansion.Value)
    End Sub
    'txtFindstr events
    Private Sub txtFindstr_LostFocus(ByVal sender As Object, ByVal e As System.EventArgs) Handles txtFindstr.LostFocus
        'Set focus on txtFindstr as long as the form is active
        If m_blnfrmViewFieldsActive Then txtFindstr.Focus()
    End Sub
    Private Sub txtFindstr_TextChanged(ByVal sender As Object, ByVal e As System.EventArgs) Handles txtFindstr.TextChanged
        If txtFindstr.Text <> "" Then
            btnFind.Enabled = True
        Else
            btnFind.Enabled = False
        End If

        'Set btnFind button state
        m_btnFindState = BTN_FIND
        btnFind.Text = "&Find"
        ToolTip.SetToolTip(btnFind, "Press Alt-F to start search")

        'Reset selection if not selecting nodes
        If Not (m_blnShiftIsDown.Value Or m_blnCtrlIsDown.Value) Then
            Tree.SelectedNode = Nothing
            ClearNodeSelection()
        End If

        'Reset status message
        txtMessage.Text = ""
    End Sub
    'Tree object events
    Private Sub Tree_BeforeExpand(ByVal sender As Object, ByVal e As System.Windows.Forms.TreeViewCancelEventArgs) Handles Tree.BeforeExpand
        'Set the flag
        m_blnTreeCollapseExpansionFlag = True

        'If this is the first time the tree is expanded then reset the font of child nodes
        If m_blnFirstExpansion.Value Then
            For i As Integer = 0 To Tree.Nodes.Count - 1
                If Tree.Nodes(i).FirstNode IsNot Nothing Then SetTreeNodesFont(m_Font, Tree.Nodes(i).FirstNode)
            Next
            m_blnFirstExpansion.Value = False
        End If
    End Sub
    Private Sub Tree_BeforeCollapse(ByVal sender As Object, ByVal e As System.Windows.Forms.TreeViewCancelEventArgs) Handles Tree.BeforeCollapse
        'Set the flag
        m_blnTreeCollapseExpansionFlag = True
    End Sub
    Private Sub Tree_MouseDown(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Tree.MouseDown
        'Set flags to show that the mouse is down.
        m_blnMouseIsDown = True
        'MouseButton not used at the moment
        m_MouseButton = e.Button
    End Sub
    Private Sub Tree_MouseUp(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Tree.MouseUp
        'Do not proceed if the mouse clic was due to a collapse expansion operation
        If m_blnTreeCollapseExpansionFlag Then
            m_blnTreeCollapseExpansionFlag = False
            Return
        End If

        'Set the flags to show that the mouse is up.
        m_blnMouseIsDown = False
        m_MouseButton = System.Windows.Forms.MouseButtons.None

        Dim LastSelectedNode, StartNode, EndNode, CurrentNode, SelectedNode As TreeNode
        Dim Comparison As Integer

        If Not (m_blnShiftIsDown.Value Or m_blnCtrlIsDown.Value) Then
            'No modifier key is pressed, m_SelectedNodesArray will contain only currently selected node
            'Reset txtFindstr string
            txtFindstr.Text = ""

            'Select node
            Tree.SelectedNode = Tree.GetNodeAt(e.Location)
            HighlightNode(Tree.SelectedNode)

            ReDim m_SelectedNodesArray(0)
            m_SelectedNodesArray(0) = Tree.SelectedNode
        Else
            'Shift key was pressed, include all nodes from last selected node up to currently selected node in m_SelectedNodesArray
            If m_blnShiftIsDown.Value Then
                Tree.SelectedNode = Tree.GetNodeAt(e.Location)

                If m_FoundNodesArray.SelectedNode > -1 And UBound(m_FoundNodesArray.Element) = UBound(m_SelectedNodesArray) Then
                    LastSelectedNode = m_FoundNodesArray.Element(m_FoundNodesArray.SelectedNode)
                Else
                    If UBound(m_SelectedNodesArray) > -1 Then
                        LastSelectedNode = m_SelectedNodesArray(UBound(m_SelectedNodesArray))
                    Else
                        'If there was no previous selection
                        LastSelectedNode = Nothing
                    End If
                End If

                Comparison = CompareTreeNodes(Tree.SelectedNode, LastSelectedNode)

                'If Nodes are equal do nothing (just update txtFindstr and txtMessage)
                If Comparison = 0 Then
                    'Reset txtFindstr string
                    txtFindstr.Text = ""

                    'Update txtMessage
                    If UBound(m_SelectedNodesArray) > 0 Then
                        txtMessage.Text = "Selected " & UBound(m_SelectedNodesArray) + 1 & " records."
                    Else
                        txtMessage.Text = ""
                    End If

                    Return
                End If

                'Else fill m_SelectedNodesArray
                'Set StartNode, EndNode, CurrentNode nodes
                If Comparison > 0 Then
                    'Check in case LastSelectedNode is Nothing
                    If LastSelectedNode Is Nothing Then
                        StartNode = Tree.SelectedNode
                    Else
                        StartNode = LastSelectedNode.NextVisibleNode
                    End If

                    EndNode = Tree.SelectedNode
                Else
                    StartNode = Tree.SelectedNode

                    'Check in case LastSelectedNode is Nothing
                    If LastSelectedNode Is Nothing Then
                        EndNode = Tree.SelectedNode
                    Else
                        EndNode = LastSelectedNode.PrevVisibleNode
                    End If
                End If
                CurrentNode = StartNode

                'Populate m_SelectedNodesArray
                'If StartNode and EndNode are equal then only one element will be inserted
                Do
                    If Not NodeIsAlreadySelected(CurrentNode) Then
                        'Highlight and update IsSelected flag
                        HighlightNode(CurrentNode)

                        'Insert node in m_SelectedNodesArray
                        ReDim Preserve m_SelectedNodesArray(UBound(m_SelectedNodesArray) + 1)
                        m_SelectedNodesArray(UBound(m_SelectedNodesArray)) = CurrentNode
                    End If

                    CurrentNode = CurrentNode.NextVisibleNode
                    'If last node was reached then exit
                    If CurrentNode Is Nothing Then Exit Do

                    'Repeat loop until CurrentNode is more than EndNode
                Loop Until CompareTreeNodes(CurrentNode, EndNode) > 0

                'Reset txtFindstr string
                txtFindstr.Text = ""
            End If

            'Control key was pressed, include currently selected node in m_SelectedNodesArray
            'unless it is already selected in which case exclude it from m_SelectedNodesArray
            If m_blnCtrlIsDown.Value Then
                SelectedNode = Tree.GetNodeAt(e.Location)

                If Not NodeIsAlreadySelected(SelectedNode) Then
                    Tree.SelectedNode = SelectedNode

                    'Highlight and update IsSelected flag
                    HighlightNode(SelectedNode)

                    'Add selected node
                    ReDim Preserve m_SelectedNodesArray(UBound(m_SelectedNodesArray) + 1)
                    m_SelectedNodesArray(UBound(m_SelectedNodesArray)) = SelectedNode
                Else
                    'Unselect selected node
                    UnSelectNode(SelectedNode)
                    UnHighlightNode(SelectedNode)
                End If

                'Reset txtFindstr string
                txtFindstr.Text = ""
            End If

            'Update txtMessage
            If UBound(m_SelectedNodesArray) > 0 Then
                txtMessage.Text = "Selected " & UBound(m_SelectedNodesArray) + 1 & " records."
            Else
                txtMessage.Text = ""
            End If
        End If
    End Sub
    Private Sub Tree_MouseMove(ByVal sender As Object, ByVal e As System.Windows.Forms.MouseEventArgs) Handles Tree.MouseMove
        If Not m_blnMouseIsDown Then
            'Ignore MouseMove if either Shift or Ctrl is down
            If m_blnShiftIsDown.Value Or m_blnCtrlIsDown.Value Then Return
        End If

        If m_blnMouseIsDown And Tree.SelectedNode IsNot Nothing Then
            'Initiate dragging
            'BuildColumnHeadersString will only pass up to MAX_NUMBER_OF_FIELDS
            Tree.DoDragDrop(BuildColumnHeadersString(), DragDropEffects.Copy)
            m_blnMouseIsDown = False
        End If
    End Sub
    Private Sub Tree_BeforeSelect(ByVal sender As Object, ByVal e As System.Windows.Forms.TreeViewCancelEventArgs) Handles Tree.BeforeSelect
        If Tree.SelectedNode Is Nothing Then Return

        'Unhighlight currently selected node unless btnNext button was clicked
        If Not m_blnNextButtonClicked Then
            'Check also that neither Shift nor Ctrl is pressed (for Tree.MouseUp events)
            If Not (m_blnShiftIsDown.Value Or m_blnCtrlIsDown.Value) Then UnHighlightNode(Tree.SelectedNode)
        End If
    End Sub
    Private Sub Tree_AfterSelect(ByVal sender As Object, ByVal e As System.Windows.Forms.TreeViewEventArgs) Handles Tree.AfterSelect
        'ClearNodeSelection only if event was raised by manual node selection
        If m_blnManualUserSelection Then
            'Check also that neither Shift nor Ctrl is pressed (for Tree.MouseUp events)
            If Not (m_blnShiftIsDown.Value Or m_blnCtrlIsDown.Value) Then
                'Finally check that btnCollapseAll was not pressed
                If Not m_blnCollapseAllClicked Then ClearNodeSelection()
            End If
        End If
    End Sub
    'Form button events
    Private Sub btnFind_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnFind.Click
        If m_btnFindState = BTN_FIND Then
            'Get total count of nodes matching the search criteria
            Dim TotalNodesFound As Integer = FindNode(Tree, txtFindstr.Text)

            If TotalNodesFound > 0 Then
                'Unhighlight currently selected node
                If Tree.SelectedNode IsNot Nothing Then UnHighlightNode(Tree.SelectedNode)

                'Set m_FoundNodesArray.SelectedNode to last element before calling btnFind_Click (with BTN_NEXT state)
                m_FoundNodesArray.SelectedNode = UBound(m_FoundNodesArray.Element)

                'Highlight each selected node
                For Each node As TreeNode In m_FoundNodesArray.Element
                    HighlightNode(node)
                Next

                'Set button btnFind new properties
                btnFind.Text = "&Next"
                m_btnFindState = BTN_NEXT
                ToolTip.SetToolTip(btnFind, "Press Alt-N to go to next record")

                'Move to first element
                btnFind_Click(sender, New System.EventArgs)

            End If
        ElseIf m_btnFindState = BTN_NEXT Then
            'Set flag for BeforeSelect event to prevent unhighlighting of a node
            m_blnNextButtonClicked = True
            'Set flag for AfterSelect event to prevent resetting of m_SelectedNodesArray
            m_blnManualUserSelection = False

            'Update m_FoundNodesArray index to next selected node (circular)
            If m_FoundNodesArray.SelectedNode = UBound(m_FoundNodesArray.Element) Then
                m_FoundNodesArray.SelectedNode = 0
            Else
                m_FoundNodesArray.SelectedNode = m_FoundNodesArray.SelectedNode + 1
            End If

            'Select Tree element
            Tree.SelectedNode = m_FoundNodesArray.Element(m_FoundNodesArray.SelectedNode)
            Tree.TopNode = Tree.SelectedNode

            'Update txtMessage to reflect currently selected node
            txtMessage.Text = "Cursor is on record " & m_FoundNodesArray.SelectedNode + 1 & " of " & UBound(m_FoundNodesArray.Element) + 1 & " records found."

            'Reset flags
            m_blnNextButtonClicked = False
            m_blnManualUserSelection = True
        End If
    End Sub
    Private Sub btnExpandAll_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnExpandAll.Click
        Dim FormerTopNode As TreeNode = Tree.TopNode
        Dim Position As Integer = GetNodeVisiblePosition(Tree.SelectedNode)
        Dim CurrentTopNode As TreeNode

        Tree.ExpandAll()

        'Set Tree.TopNode
        'its value depends on whereas Tree.SelectedNode is Nothing or not
        If Tree.SelectedNode Is Nothing Then
            Tree.TopNode = FormerTopNode
        Else
            CurrentTopNode = Tree.SelectedNode
            If Position > 0 And CurrentTopNode IsNot Nothing Then
                For i As Integer = 1 To Position - 1
                    'Get CurrentTopNode
                    If CurrentTopNode.PrevVisibleNode Is Nothing Then Exit For
                    CurrentTopNode = CurrentTopNode.PrevVisibleNode
                Next
            End If
            Tree.TopNode = CurrentTopNode
        End If

        'Reset this flag which is set automatically during the operation of Tree.CollapseAll
        m_blnTreeCollapseExpansionFlag = False

        'Set buttons state
        btnExpandAll.Enabled = False
        btnCollapseAll.Enabled = True
    End Sub
    Private Sub btnCollapseAll_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnCollapseAll.Click
        Dim FormerTopNode As TreeNode = Tree.TopNode
        Dim FormerSelectedNode As TreeNode = Tree.SelectedNode
        Dim Position As Integer = GetNodeVisiblePosition(Tree.SelectedNode)
        Dim CurrentTopNode As TreeNode

        'Tree.CollapseAll sets Tree.SelectedNode to Nothing if the former Tree.SelectedNode was at the root level
        'or to the root parent of the former Tree.SelectedNode if it was below the root level
        'this generates the events Tree.BeforeSelect and Tree.AfterSelect
        'Set this flag in order to prevent ClearNodeSelection
        m_blnCollapseAllClicked = True

        Tree.CollapseAll()

        Tree.SelectedNode = FormerSelectedNode

        'Set Tree.TopNode
        'its value depends on whereas Tree.SelectedNode is Nothing or not
        If Tree.SelectedNode Is Nothing Then
            Tree.TopNode = FormerTopNode
        Else
            CurrentTopNode = Tree.SelectedNode
            If Position > 0 And CurrentTopNode IsNot Nothing Then
                For i As Integer = 1 To Position - 1
                    'Get CurrentTopNode
                    If CurrentTopNode.PrevVisibleNode Is Nothing Then Exit For
                    CurrentTopNode = CurrentTopNode.PrevVisibleNode
                Next
            End If
            Tree.TopNode = CurrentTopNode
        End If

        'Reset this flag which is set automatically during the operation of Tree.CollapseAll
        m_blnTreeCollapseExpansionFlag = False

        'Reset m_blnCollapseAll flag
        m_blnCollapseAllClicked = False

        'Set buttons state
        btnCollapseAll.Enabled = False
        btnExpandAll.Enabled = True
    End Sub
    Private Sub btnSort_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnSort.Click
        If m_btnSortState = BTN_SORT Then
            'Reset node selection and search criteria
            txtFindstr.Text = ""
            Tree.SelectedNode = Nothing

            'Set bold font prior to sort the TreeView structure so that highlighted elements display correctly
            SetTreeNodesFont(m_BoldFont)

            'There is no need to reset m_Font to child nodes at this point 
            If m_blnFirstExpansion.Value Then m_blnFirstExpansion.Value = False

            'Sort based on TreeNode.Text criteria
            Tree.Sort()

            'If NodeTag.SortedIndex is not set then initialize it
            If CType(Tree.Nodes(0).Tag, NodeTag).SortedIndex <= 0 Then
                SetSortedIndex()
            End If

            'Reset the TreeView font
            SetTreeNodesFont(m_Font)

            Tree.CollapseAll()

            'Select first node in the Tree structure
            Tree.SelectedNode = Tree.Nodes(0)
            HighlightNode(Tree.SelectedNode)

            'Fill SelectedNodesArray with current selection
            ReDim m_SelectedNodesArray(0)
            m_SelectedNodesArray(0) = Tree.SelectedNode

            'Set button btnSort new properties
            btnSort.Text = "Unsor&t"
            m_btnSortState = BTN_UNSORT
            ToolTip.SetToolTip(btnSort, "Press Alt-T to unsort tree")
            'Set expand and collapse buttons state
            If m_blnWithDepth Then btnExpandAll.Enabled = True
            btnCollapseAll.Enabled = False

        ElseIf m_btnSortState = BTN_UNSORT Then
            'Reset node selection and search criteria
            txtFindstr.Text = ""
            Tree.SelectedNode = Nothing

            'Set bold font prior to sort the TreeView structure so that highlighted elements display correctly
            SetTreeNodesFont(m_BoldFont)

            'Sort using the personalized sorter (based on NodeTag.UnsortedIndex property) 
            'and then restore to the default sorter
            Tree.TreeViewNodeSorter = TagSorter.DefaultSorter
            Tree.Sort()
            Tree.TreeViewNodeSorter = m_TreeSorter

            'Reset the TreeView font
            SetTreeNodesFont(m_Font)

            Tree.CollapseAll()

            'Select first node in the Tree structure
            Tree.SelectedNode = Tree.Nodes(0)
            HighlightNode(Tree.SelectedNode)

            'Fill SelectedNodesArray with current selection
            ReDim m_SelectedNodesArray(0)
            m_SelectedNodesArray(0) = Tree.SelectedNode

            'Set button btnSort new properties
            btnSort.Text = "Sor&t"
            m_btnSortState = BTN_SORT
            ToolTip.SetToolTip(btnSort, "Press Alt-T to sort tree")
            'Set expand and collapse buttons state
            If m_blnWithDepth Then btnExpandAll.Enabled = True
            btnCollapseAll.Enabled = False
        End If

    End Sub
    Private Sub btnSave_Click(ByVal sender As System.Object, ByVal e As System.EventArgs) Handles btnSave.Click
        'Show SaveFileDialog to allow user enter path and filename
        SaveFileDialog.Filter = "csv Files|*.csv"
        SaveFileDialog.Title = "Save csv File"
        If SaveFileDialog.ShowDialog <> System.Windows.Forms.DialogResult.OK Then Return

        Try
            'Open File for writing
            Dim FileNumber As Integer = FreeFile()
            FileOpen(FileNumber, SaveFileDialog.FileName, OpenMode.Output)

            'Print headers
            PrintLine(FileNumber, "Field Name,Value")

            Dim TempNode As TreeNode
            'Save only selected items if there are more than one
            If UBound(m_SelectedNodesArray) >= 1 Then
                For i As Integer = 0 To UBound(m_SelectedNodesArray)
                    TempNode = m_SelectedNodesArray(i)
                    PrintLine(FileNumber, GetText(TempNode))
                Next
            Else
                'Save the whole Tree
                'Select first element in Tree
                TempNode = Tree.Nodes(0)
                'Print every visible node to file
                Do Until TempNode Is Nothing
                    PrintLine(FileNumber, GetText(TempNode))
                    TempNode = TempNode.NextVisibleNode
                Loop
            End If

            FileClose(FileNumber)
        Catch ex As Exception
            Debug.Print("File " & SaveFileDialog.FileName & " cannot be opened for write access.")
            MsgBox("File " & SaveFileDialog.FileName & " cannot be opened for write access.")
        End Try
    End Sub
#End Region
End Class
