///////////////////////////////////////////////////////////////////////////////////////////////
// Datatec all rights reserved
// drvapi.cpp : test module of the C++ DcApi library
///////////////////////////////////////////////////////////////////////////////////////////////

#include <stdio.h>
#include <time.h>
#include <stdarg.h>

#define GLOBALS_HERE
#include "drvapi.h"

static void     L_Log_Print           (char *format, ... );

static int L_counter;

int CDrvApi::Set_Verbose_Mode (void)
{
    m_verbose_mode = !m_verbose_mode;
    return m_verbose_mode;
}

int CDrvApi::Verbouse_Mode (void)
{
    return m_verbose_mode;
}
// Receiving message callback
void CDrvApi::OnMessageReceived (char *message_type, unsigned int message_num)
{
  L_Log_Print ("OnMessageReceived : %s number : %u\n", message_type, message_num);  
  
  // Check if verbose mode feature is activated
  if (m_verbose_mode)
    DisplayMessageContent (message_type, message_num);
}

// TODO method Connection stablished
// by now, only report process is done
void CDrvApi::OnConnection    (int result_code)
{
  if (result_code != 0)
  {
    L_Log_Print ("Connection try failed, result code : %d\n", result_code);
  }
  else
    L_Log_Print ("Connection established.\n");
}

// Disconnection occurred
void CDrvApi::OnDisconnection (int reason_code, char * reason_text)
{
  L_Log_Print ("Disconnection occurred, reason code %d, reason text %s\n", reason_code, reason_text);
}

// Error ocurred
void CDrvApi::OnError     (int error_code, char * error_text, int next_action_code, int next_action_data)
{
  L_Log_Print ("Error %d, %s. Next sugested action code %d-%d.\n",  error_code,
                                                                    error_text,
                                                                    next_action_code,
                                                                    next_action_data);
}

// Low logging status changing
void CDrvApi::OnStatusChanged   (int status_code, char * status_text)
{
  L_Log_Print ("Status changed %d, %s\n", status_code, status_text);
}

/////////
// from here : sample code to do something
/////////

// sample code to display all the fields of a message
// previously received
void CDrvApi::DisplayMessageContent (char *message_type, unsigned int message_num)
{
  char field_data[4096];
  char field_name[1024];
  int  field_type;
  int  i;
  int  i_op_field_value;

  // indicate to the API what message we are going to view
  // call the same way as many times as it is required till the message is no more
  // required. (to free the internal buffer in the API call FinishMessageProcess)
  if (TakeMessage (message_type, message_num))
  {
    L_Log_Print ("DisplayMessageContent error TakeMessage. Message %s. Number: %u\n", message_type, message_num);  
    return;
  }

  *field_data = 0;
  i_op_field_value = 0;

  // All global tables (defined with table_code parameter in config file)
  // must validate the operation field to avoid issues with messages
  // without valid content.
  for (i=0; i<DAPI_GLOBAL_TABLE_SIZE; ++i)
  {
    // DAPI_Global_Table_Names[] contains the strings of the global tables definitions.
    // "Posturas_Vigentes", "Resumenes_Vigentes", "Grupos_Vigentes", "Mercados_Renta_Fija_Vigentes", "Hechos_Vigentes", "Indices_Vigentes"
    if (!_stricmp(message_type, DAPI_Global_Table_Names [i][Glb_Old_Names])  ||
        !_stricmp(message_type, DAPI_Global_Table_Names [i][Glb_DFN_Names])  ||
        !_stricmp(message_type, DAPI_Global_Table_Names [i][Glb_English_Names]))
    {
      // For global tables, the operation field must be one of the following ones.
      field_type = GetField ("i_op", field_data);
      if (!field_type)
      {
        L_Log_Print ("DisplayMessageContent error getting i_op field. Message %s. Number : %d\n", message_type, message_num);  
        return;
      }
      
      // i_op field can be sent in upper or lower case.
      i_op_field_value = toupper(*field_data);
      switch (i_op_field_value)
      {
      case 'I': // Insert
      case 'C': // Copy
      case 'A': // Update
        // Global table can be processed.
        break;
      default:
        // Delete (B) is a special case. Record is sent with an ID but without valid content.
        // All other values must not be taken into account to be processed.
        L_Log_Print ("DisplayMessageContent. Global table SKIPPED. i_op: %c. Message %s. Number : %d\n", i_op_field_value, message_type, message_num);
        return;
      }

      break;
    }
  }

  // Get all the fields in the message, for testing purpose
  // only the required fields should be taken like   the next sentence
  // GetFieldBSTR ("TRADE_IDENTIFICATION", trade_id);

  field_name[0] = 0; // To take the first at the beggining
  for (;;)
  {
    field_type = GetNextField (field_name, field_data);
    if (!field_type)
      break;

    L_Log_Print ("  Field %s :", field_name);
    switch (field_type)
    {
    case FLD_SHORT :
      L_Log_Print (" short %hd\n", *(short *)field_data);
      break;
    case FLD_UNSIGNED_SHORT :
      L_Log_Print (" unsigned short %hu\n", *(unsigned short *)field_data);
      break;
    case FLD_UNSIGNED_INT:
      L_Log_Print (" unsigned int %u\n", *(unsigned int*)field_data);
      break;
    case FLD_ALPHA :
    case FLD_ALPHA_TYPE_STRING :
      L_Log_Print (" alpha %s\n", field_data);
      break;
    case FLD_NUM_ASCII_DOUBLE :
      L_Log_Print (" num-point %s\n", field_data);
      break;
    case FLD_NUM_ASCII_INT :
      L_Log_Print (" num-ascii %s\n", field_data);
      break;
    case FLD_INT :
      L_Log_Print (" int %d\n", *(int*)field_data);
      break;
    }
  }
}

// Sample code to send a message
void CDrvApi::DoSend (void)
{
  // begin the record
  InitMessage (L_GuibosType);
  FillCommonFields();
  // Fill buyer and seller fields.
  FillBuyerFields();
  FillSellerFields();
  SetFieldBSTR(L_GuibosType, (char *)"flag_side_changed", (char *)"A");
  // send the message
  SendMessage(L_GuibosType);
}

void CDrvApi::SendTestMessage (void)
{
  // begin the record
  InitMessage (L_Orden_TransaccionType);

  // Llenar los campos del mensaje a ser enviado.
  FillBodyMsg();
  
  // send the message
  SendMessage(L_Orden_TransaccionType);
}

// sample code to process a message 
// received message, and not processed untill now
void CDrvApi::DoFinishMessageProcess (char * message_type, unsigned int message_num)
{
  if (TakeMessage (message_type, message_num))
    return;

  InitMessage ((char*)""); // "" => initiates the same taken messages with it's content
                            // this avoids to construct all the message fields

  // send back the same message
  SendMessage(message_type);
}

// sample code to signal a message as processed
// no further processed will be done with the content of the message
void CDrvApi::FlagMessage (char * message_type, unsigned int message_num)
{
  FinishMessageProcess (message_type, message_num); // the message is signaled and deleted
}

// sample code to fill some fields of a message previously defined with InitMessage
void CDrvApi::FillBuyerFields(void)
{
  char date_str[16];
  char time_str[8];

  // Fill these fields also in case of using a blank message for testing.
  SetFieldBSTR(L_GuibosType, (char*)"deal_type_buyer_bank",           (char*)"E");
  SetFieldBSTR(L_GuibosType, (char*)"buyer_bank",                     (char*)"AAALUK2LXXX");
  SetFieldBSTR(L_GuibosType, (char*)"buyer_dealer",                   (char*)"AAALUK2LXXX001");
  SetFieldBSTR(L_GuibosType, (char*)"code_buyer_via",                 (char*)"BBBGUKLC");
  SetFieldBSTR(L_GuibosType, (char*)"broker_buyer",                   (char*)"LGEU543");
  SetFieldBSTR(L_GuibosType, (char*)"buyer_side_net_brokerage",       (char*)"N");
  SetFieldBSTR(L_GuibosType, (char*)"flag_buyer_aggresor",            (char*)"");
  SetFieldBSTR(L_GuibosType, (char*)"buyer_side_status_error",        (char*)"");//<=

  GetCurrentDateTime(date_str, time_str);

  // These are the fields that should be filled when the message is received from the DS.
  SetFieldBSTR(L_GuibosType, (char*)"buyer_published_field",          (char*)"N");//<=
  SetFieldBSTR(L_GuibosType, (char*)"buyer_side_drn",                 (char*)"KKPLZNH1");//<=
  SetFieldBSTR(L_GuibosType, (char*)"buyer_side_version",             (char*)"1.234");//<=
  SetFieldBSTR(L_GuibosType, (char*)"buyer_side_hour_entered",        (char*)time_str);//<=
  SetFieldBSTR(L_GuibosType, (char*)"buyer_side_date_entered",        (char*)date_str);//<=
  SetFieldBSTR(L_GuibosType, (char*)"flag_side_changed",              (char*)"B");//"B" - Buyer,  "S" - Seller, "A" - both sides
}

// sample code to fill some fields of a message previously defined with InitMessage
void CDrvApi::FillSellerFields(void)
{
  char date_str[16];
  char time_str[8];

  SetFieldBSTR(L_GuibosType, (char*)"deal_type_seller_bank",          (char*)"E");
  SetFieldBSTR(L_GuibosType, (char*)"seller_bank",                    (char*)"BACYUS3LXXX");
  SetFieldBSTR(L_GuibosType, (char*)"seller_dealer",                  (char*)"BBBLUK2LXXX001");
  SetFieldBSTR(L_GuibosType, (char*)"code_seller_via",                (char*)"GHICUSOP");
  SetFieldBSTR(L_GuibosType, (char*)"broker_seller",                  (char*)"NAPI543");
  SetFieldBSTR(L_GuibosType, (char*)"seller_side_net_brokerage",      (char*)"N");
  SetFieldBSTR(L_GuibosType, (char*)"flag_seller_aggresor",           (char*)"");
  SetFieldBSTR(L_GuibosType, (char*)"seller_side_status_error",       (char*)"");//<=

  GetCurrentDateTime(date_str, time_str);

  // These are the fields that should be filled when the message is received from the DS.
  SetFieldBSTR(L_GuibosType, (char*)"seller_published_field",        (char*) "N");//<=
  SetFieldBSTR(L_GuibosType, (char*)"seller_side_drn",               (char*) "LLQMAOI2");//<=
  SetFieldBSTR(L_GuibosType, (char*)"seller_side_version",           (char*) "1.235");//<=
  SetFieldBSTR(L_GuibosType, (char*)"seller_side_hour_entered",      (char*) time_str);//<=
  SetFieldBSTR(L_GuibosType, (char*)"seller_side_date_entered",      (char*) date_str);//<=
  SetFieldBSTR(L_GuibosType, (char*)"flag_side_changed",             (char*) "S");//"B" - Buyer,  "S" - Seller, "A" - both sides
}

// sample code to fill fields of an empty message
void CDrvApi::FillCommonFields (void)
{
  char trade_id[100];
  static unsigned int counter;

  // assemble a record with all the fields
  // in this case all the fields are set with a sample value in order
  // to fill the message. Usually the fields has the same
  // value as the previously received message excepting thos marked with //<-
  SetFieldBSTR(L_GuibosType, (char*)"bandera",                        (char*)"A");
  SetFieldBSTR(L_GuibosType, (char*)"instrument",                     (char*)"SWAP");
  SetFieldBSTR(L_GuibosType, (char*)"currency_pair",                  (char*)"EURUSD");
  SetFieldBSTR(L_GuibosType, (char*)"deal_time",                      (char*)"181025");
  SetFieldBSTR(L_GuibosType, (char*)"deal_date",                      (char*)"20011114");
  snprintf (trade_id, 100,   "%015d", ++counter);
  SetFieldBSTR(L_GuibosType, (char*)"trade_identification",           trade_id);
  SetFieldBSTR(L_GuibosType, (char*)"instance_trade",                 (char*)"1234");
  SetFieldBSTR(L_GuibosType, (char*)"points",                         (char*)"-27.1");
  SetFieldBSTR(L_GuibosType, (char*)"currency_code_dealt",            (char*)"EUR");
  SetFieldBSTR(L_GuibosType, (char*)"value_date_near",                (char*)"20011116");
  SetFieldBSTR(L_GuibosType, (char*)"amount_near",                    (char*)"100000(char*)00000.00");
  SetFieldBSTR(L_GuibosType, (char*)"currency_code_nondealt",         (char*)"EUR");
  SetFieldBSTR(L_GuibosType, (char*)"rate",                           (char*)".5678");
  SetFieldBSTR(L_GuibosType, (char*)"value_date_far",                 (char*)"20101116");
  SetFieldBSTR(L_GuibosType, (char*)"amount_far",                     (char*)"10000000000.00");
  SetFieldBSTR(L_GuibosType, (char*)"date_register",                  (char*)"20011114");
  SetFieldBSTR(L_GuibosType, (char*)"time_register",                  (char*)"182020");
  SetFieldBSTR(L_GuibosType, (char*)"canceled_amend",                 (char*)"C");
  SetFieldBSTR(L_GuibosType, (char*)"text_term",                      (char*)"1Y");
  SetFieldBSTR(L_GuibosType, (char*)"mar_date",                       (char*)"20101110");
  SetFieldBSTR(L_GuibosType, (char*)"filler",                         (char*)"");
  SetFieldBSTR(L_GuibosType, (char*)"key_used_datatec",               (char*)"");
  SetFieldUINT(L_GuibosType, (char*)"part_1",                                 0);
  SetFieldBSTR(L_GuibosType, (char*)"part_2",                         (char*)"");
  SetFieldBSTR(L_GuibosType, (char*)"part_3",                         (char*)"");
  SetFieldBSTR(L_GuibosType, (char*)"part_4",                         (char*)"");
  // Seller side DRN (Near End) ???
  // Buyer status flag ???
  // Seller status flag ???
}

void CDrvApi::FillBodyMsg (void)
{
  char counter_str[30];

  // Debe ser definido con 71 (alfanumerico de 2)
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"clase_mensaje", (char*)"71");
  
  // S - Le indica al DS que este mensaje no se debe grabar en el log de los destinatarios 
  // del mensaje (alfanumerico de 1)
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"no_guardar_log_destinatarios", (char*)"S");

  // Es el tipo de operacin (alfanumerico de 1), sus posibles valores son:
  //   A = Alta, 
  //   B = Baja
  //   M = Modificacin
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"bandera", (char*)"A");

  // El numero del mercado a que corresponde este mensaje (alfanumerico de 3)
  // Los posibles cdigos de mercado son: 69, 140  141
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"mercado", (char*)"69");

  // La fecha de la transaccion. Formato: AAAAMMDD. (alfanumerico de 8)
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"fecha_postura", (char*)"20090601");

  // Es la hora_hecho, es decir es la hora de la ultima operacion sobre la postura
  // (alfanumerico de 6)
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"hora_operacion", (char*)"124000");

  // Folio.  El numero de papeles que estan asignados (entero doble)
  L_counter++;
  sprintf (counter_str, "%d", L_counter);
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"numero_asignado", (char*)counter_str);

  // Cdigo alterno del grupo, usado para ingresar una nueva lnea (alfanumerico de 1)
  // Los posibles valores son: C  V
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"orden_operacion", (char*)"C");

  // Es el tipo de operacin (alfanumerico de 2). Los posibles valores son:
  //    R = Reporto
  //    D = Directa
  //    DR  = Directa y repoto
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"tipo_operacion", (char*)"D");

  // Nombre del ttulo (alfanumerico de 10). Ejemplo: JL09, DC10, CTSI, G2I, etc.
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"texto_term", (char*)"JL09");

  // Horas hasta la liquidacin (entero sin signo), posibles valores:
  //    0 = Mismo da (MD)
  //    24, 48, 72, 96 horas
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"liquidacion", (char*)"24");

  // Concertacion (entero sin signo). Los posibles valores son:
  //    0 = Orden en firme
  //    1 = Cama
  //    2 = Ronda
  //    3 = Chinga
  //    4 = Montos
  //    5 = Ponderada
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"concertacion", (char*)"1");

  // Plazo de la transaccin  plazo inferior cuando es un rango, ejemplo (60-360) 
  // (entero)
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"xdias", (char*)"60");

  // Plazo superior cuando es un rango, ejemplo (60-360) 
  // (entero si signo)
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"dias", (char*)"360");

  // Cdigo de la institucin compradora (alfanumerico de 10). Ejmplo: BANAMEX
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"referencia", (char*)"BANAMEX");

  // Nmero asociado a la institucin compradora (entero sin signo). Ejemplo: 26
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"numero_sucursal_postura", (char*)"26");

  // Nmero asociado a la institucin vendedora (entero sin signo). Ejemplo: 54
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"numero_sucursal_postura_1", (char*)"54");

  // ' Precio de la transaccin (numerico 13 enteros, 4 decimales). Ejemplo: 7.3, 5.21, etc
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"precio", (char*)"5.21");
   
  // Monto de la transaccin (numerico 13 enteros y 2 decimales)
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"numero_transado", (char*)"150.39");

  // Los posibles valores son: 
  //    C = orden/pedido de compra
  //    V = orden/pedido de venta
  //    (alfanumerico de 1)
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"o_d", (char*)"C");

  // Indica el agresor (alfanumerico de 3), los posibles valores son: 
  //    CO = Comprador fue el agresor,  
  //    CE = Vendedor fue el agresor
  SetFieldBSTR(L_Orden_TransaccionType, (char*)"entidad_publica_3c", (char*)"CE");
}

//////////////////////////////
// EXTRA code to help owrking

void GetCurrentDateTime (char *date_str, char *time_str)
{
  unsigned long cur_time = 0;
  struct tm today;

  cur_time = (unsigned long) time(NULL);
  memcpy(&today, localtime((time_t *)&cur_time), sizeof(struct tm));
  sprintf(date_str, "%4d%2d%2d", today.tm_year, today.tm_mon+1, today.tm_mday);
  sprintf(time_str, "%2d%2d%2d", today.tm_hour, today.tm_min, today.tm_sec);
}

void L_Log_Print (char *format, ... )
{
  static FILE *log_file;

  // check if the log file is opened
  if (!log_file)
  {
    // no, open the file
    log_file = fopen ("dapi_c_log", "wt");

    // if log is still not possible terminate now
    if (!log_file)
      return;
  }

  va_list marker, marker2;
  
  va_start (marker,  format);     /* Initialize variable arguments. */
  va_start (marker2, format);

  // printf
  vfprintf (log_file, format, marker);
  vprintf  (format, marker2);

  va_end(marker);
  va_end(marker2);
}

/////////////////////////////////////////////////////////////////////////////////////////////////////////////

